import os

from fastapp_template._zrb.config import (
    ACTIVATE_VENV_SCRIPT,
    APP_DIR,
    MICROSERVICES_ENV_VARS,
    MONOLITH_ENV_VARS,
)

from zrb import Cmd, CmdTask, EnvFile, EnvMap, StrInput, Task


def create_migration(name: str, module: str) -> Task:
    return CmdTask(
        name=f"create-app-name-{name}-migration",
        description=f"🧩 Create App Name {name.capitalize()} DB migration",
        input=StrInput(
            name="message",
            description="Migration message",
            prompt="Migration message",
            allow_empty=False,
        ),
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            EnvMap(
                vars={
                    "APP_DB_URL": f"sqlite:///{APP_DIR}/.migration.{module}.db",
                    "APP_NAME_MODULES": f"{module}",
                }
            ),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            f"cd {os.path.join(APP_DIR, 'module', module)}",
            "alembic upgrade head",
            Cmd(
                "alembic revision --autogenerate -m {double_quote(ctx.input.message)}",
                auto_render=True,
            ),
        ],
        render_cmd=False,
        retries=2,
    )


def migrate_module(name: str, module: str, as_microservices: bool) -> Task:
    env_vars = MICROSERVICES_ENV_VARS if as_microservices else MONOLITH_ENV_VARS
    return CmdTask(
        name=(
            f"migrate-app-name-{name}"
            if as_microservices
            else f"migrate-{name}-on-monolith"
        ),
        description=f"🧩 Run App Name {name.capitalize()} DB migration",
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            EnvMap(
                vars={
                    **env_vars,
                    "APP_NAME_MODULES": f"{module}",
                }
            ),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            f"cd {os.path.join(APP_DIR, 'module', module)}",
            "alembic upgrade head",
        ],
        render_cmd=False,
        retries=2,
    )


def run_microservice(name: str, port: int, module: str) -> Task:
    return CmdTask(
        name=f"run-app-name-{name}",
        description=f"🧩 Run App Name {name.capitalize()}",
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            EnvMap(
                vars={
                    **MICROSERVICES_ENV_VARS,
                    "APP_NAME_PORT": f"{port}",
                    "APP_NAME_MODULES": f"{module}",
                }
            ),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            'fastapi dev main.py --port "${APP_NAME_PORT}"',
        ],
        render_cmd=False,
        retries=2,
    )



import datetime
import os

from zrb.session_state_log.session_state_log import SessionStateLog, SessionStateLogList
from zrb.session_state_logger.any_session_state_logger import AnySessionStateLogger


class FileSessionStateLogger(AnySessionStateLogger):

    def __init__(self, session_log_dir: str):
        self._session_log_dir = session_log_dir

    def write(self, session_log: SessionStateLog):
        session_file_path = self._get_session_file_path(session_log.name)
        session_dir_path = os.path.dirname(session_file_path)
        if not os.path.isdir(session_dir_path):
            os.makedirs(session_dir_path, exist_ok=True)
        with open(session_file_path, "w") as f:
            f.write(session_log.model_dump_json())
        start_time = session_log.start_time
        if start_time == "":
            return
        timeline_dir_path = self._get_timeline_dir_path(session_log)
        os.makedirs(timeline_dir_path, exist_ok=True)
        with open(os.path.join(timeline_dir_path, session_log.name), "w"):
            pass

    def read(self, session_name: str) -> SessionStateLog:
        session_file_path = self._get_session_file_path(session_name)
        with open(session_file_path, "r") as f:
            return SessionStateLog.model_validate_json(f.read())

    def list(
        self,
        task_path: list[str],
        min_start_time: datetime.datetime,
        max_start_time: datetime.datetime,
        page: int = 0,
        limit: int = 10,
    ) -> SessionStateLogList:
        matching_sessions = []
        # Traverse the timeline directory and filter sessions
        timeline_dir = os.path.join(self._session_log_dir, "_timeline", *task_path)
        if not os.path.exists(timeline_dir):
            return {"total": 0, "data": []}
        for root, _, files in os.walk(timeline_dir):
            for file_name in files:
                session_name = os.path.splitext(file_name)[0]
                # Read the session and retrieve start time
                session_log = self.read(session_name)
                start_time = self._get_start_time(session_log)
                # Filter sessions based on start time
                if start_time and min_start_time <= start_time <= max_start_time:
                    matching_sessions.append((start_time, session_log))
        # Sort sessions by start time, descending
        matching_sessions.sort(key=lambda x: x[0], reverse=True)
        total = len(matching_sessions)
        # Apply pagination
        start_index = page * limit
        end_index = start_index + limit
        paginated_sessions = matching_sessions[start_index:end_index]
        # Extract session logs from the sorted list of tuples
        data = [session_log for _, session_log in paginated_sessions]
        return SessionStateLogList(total=total, data=data)

    def _get_session_file_path(self, session_name: str) -> str:
        return os.path.join(self._session_log_dir, f"{session_name}.json")

    def _get_timeline_dir_path(self, session_log: SessionStateLog) -> str:
        start_time = self._get_start_time(session_log)
        year = start_time.year
        month = start_time.month
        day = start_time.day
        hour = start_time.hour
        minute = start_time.minute
        second = start_time.second
        paths = session_log.path + [
            f"{year}",
            f"{month}",
            f"{day}",
            f"{hour}",
            f"{minute}",
            f"{second}",
        ]
        return os.path.join(self._session_log_dir, "_timeline", *paths)

    def _get_start_time(self, session_log: SessionStateLog) -> datetime.datetime:
        return datetime.datetime.strptime(
            session_log.start_time, "%Y-%m-%d %H:%M:%S.%f"
        )



import libcst as cst


class ClassCodeAdder(cst.CSTTransformer):
    def __init__(self, class_name: str, new_code: str):
        self.class_name = class_name
        self.new_code = cst.parse_statement(new_code)
        self.class_found = False

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        # Check if this is the target class
        if original_node.name.value == self.class_name:
            self.class_found = True
            # Add the method to the class body
            new_body = updated_node.body.with_changes(
                body=updated_node.body.body + (self.new_code,)
            )
            return updated_node.with_changes(body=new_body)
        return updated_node


def add_code_to_class(original_code: str, class_name: str, method_code: str) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize transformer with the class name and method code
    transformer = ClassCodeAdder(class_name, method_code)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Check if the class was found
    if not transformer.class_found:
        raise ValueError(f"Class {class_name} not found in the provided code.")
    # Return the modified code
    return modified_module.code



import json
import os
from collections.abc import Callable
from typing import Any

from pydantic import BaseModel

from zrb.attr.type import StrAttr
from zrb.config import LLM_MODEL, LLM_SYSTEM_PROMPT
from zrb.context.any_context import AnyContext
from zrb.context.any_shared_context import AnySharedContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_str_attr
from zrb.util.cli.style import stylize_faint
from zrb.util.llm.tool import callable_to_tool_schema

ListOfDict = list[dict[str, Any]]


class AdditionalTool(BaseModel):
    fn: Callable
    name: str | None
    description: str | None


def scratchpad(thought: str) -> str:
    """Use this tool to note your thought and planning"""
    return thought


class LLMTask(BaseTask):

    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        model: StrAttr | None = LLM_MODEL,
        render_model: bool = True,
        system_prompt: StrAttr | None = LLM_SYSTEM_PROMPT,
        render_system_prompt: bool = True,
        message: StrAttr | None = None,
        tools: (
            dict[str, Callable] | Callable[[AnySharedContext], dict[str, Callable]]
        ) = {},
        history: ListOfDict | Callable[[AnySharedContext], ListOfDict] = [],
        history_file: StrAttr | None = None,
        render_history_file: bool = True,
        model_kwargs: (
            dict[str, Any] | Callable[[AnySharedContext], dict[str, Any]]
        ) = {},
        execute_condition: bool | str | Callable[[AnySharedContext], bool] = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
        )
        self._model = model
        self._render_model = render_model
        self._model_kwargs = model_kwargs
        self._system_prompt = system_prompt
        self._render_system_prompt = render_system_prompt
        self._message = message
        self._tools = tools
        self._history = history
        self._history_file = history_file
        self._render_history_file = render_history_file
        self._additional_tools: list[AdditionalTool] = []

    def add_tool(
        self, tool: Callable, name: str | None = None, description: str | None = None
    ):
        self._additional_tools.append(
            AdditionalTool(fn=tool, name=name, description=description)
        )

    async def _exec_action(self, ctx: AnyContext) -> Any:
        from litellm import acompletion, supports_function_calling

        model = self._get_model(ctx)
        try:
            allow_function_call = supports_function_calling(model=model)
        except Exception:
            allow_function_call = False
        model_kwargs = self._get_model_kwargs(ctx)
        ctx.log_debug("MODEL KWARGS", model_kwargs)
        system_prompt = self._get_system_prompt(ctx)
        ctx.log_debug("SYSTEM PROMPT", system_prompt)
        history = self._get_history(ctx)
        ctx.log_debug("HISTORY PROMPT", history)
        user_message = {"role": "user", "content": self._get_message(ctx)}
        ctx.print(stylize_faint(f"{user_message}"))
        messages = history + [user_message]
        available_tools = self._get_tools(ctx)
        available_tools["scratchpad"] = scratchpad
        if allow_function_call:
            tool_schema = [
                callable_to_tool_schema(tool, name)
                for name, tool in available_tools.items()
            ]
            for additional_tool in self._additional_tools:
                fn = additional_tool.fn
                tool_name = additional_tool.name or fn.__name__
                tool_description = additional_tool.description
                available_tools[tool_name] = additional_tool.fn
                tool_schema.append(
                    callable_to_tool_schema(
                        fn, name=tool_name, description=tool_description
                    )
                )
            model_kwargs["tools"] = tool_schema
            ctx.log_debug("TOOL SCHEMA", tool_schema)
        history_file = self._get_history_file(ctx)
        while True:
            response = await acompletion(
                model=model,
                messages=[{"role": "system", "content": system_prompt}] + messages,
                **model_kwargs,
            )
            response_message = response.choices[0].message
            ctx.print(stylize_faint(f"{response_message.to_dict()}"))
            messages.append(response_message.to_dict())
            tool_calls = response_message.tool_calls
            if tool_calls:
                # noqa Reference: https://docs.litellm.ai/docs/completion/function_call#full-code---parallel-function-calling-with-gpt-35-turbo-1106
                for tool_call in tool_calls:
                    function_name = tool_call.function.name
                    function_to_call = available_tools[function_name]
                    function_kwargs = json.loads(tool_call.function.arguments)
                    function_response = function_to_call(**function_kwargs)
                    tool_call_message = {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": function_name,
                        "content": function_response,
                    }
                    ctx.print(stylize_faint(f"{tool_call_message}"))
                    messages.append(tool_call_message)
                continue
            if history_file != "":
                os.makedirs(os.path.dirname(history_file), exist_ok=True)
                with open(history_file, "w") as f:
                    f.write(json.dumps(messages, indent=2))
            return response_message.content

    def _get_model(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._model, "ollama_chat/llama3.1", auto_render=self._render_model
        )

    def _get_system_prompt(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx,
            self._system_prompt,
            "You are a helpful assistant",
            auto_render=self._render_system_prompt,
        )

    def _get_message(self, ctx: AnyContext) -> str:
        return get_str_attr(ctx, self._message, "How are you?", auto_render=True)

    def _get_model_kwargs(self, ctx: AnyContext) -> dict[str, Callable]:
        if callable(self._model_kwargs):
            return self._model_kwargs(ctx)
        return {**self._model_kwargs}

    def _get_tools(self, ctx: AnyContext) -> dict[str, Callable]:
        if callable(self._tools):
            return self._tools(ctx)
        return self._tools

    def _get_history(self, ctx: AnyContext) -> ListOfDict:
        if callable(self._history):
            return self._history(ctx)
        history_file = self._get_history_file(ctx)
        if (
            len(self._history) == 0
            and history_file != ""
            and os.path.isfile(history_file)
        ):
            with open(history_file, "r") as f:
                return json.loads(f.read())
        return self._history

    def _get_history_file(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._history_file, "", auto_render=self._render_history_file
        )



from fastapp_template.config import APP_AUTH_BASE_URL
from fastapp_template.module.auth.client.base_client import BaseClient
from fastapp_template.module.auth.service.user.usecase import user_usecase


class APIClient(user_usecase.as_api_client(base_url=APP_AUTH_BASE_URL), BaseClient):
    pass



import fnmatch
import re
from collections.abc import Callable

from zrb.content_transformer.any_content_transformer import AnyContentTransformer
from zrb.context.any_context import AnyContext


class ContentTransformer(AnyContentTransformer):

    def __init__(
        self,
        match: list[str] | str | Callable[[AnyContext, str], bool],
        transform: dict[str, str] | Callable[[AnyContext, str], str],
        auto_render: bool = True,
    ):
        self._match = match
        self._transform_file = transform
        self._auto_render = auto_render

    def match(self, ctx: AnyContext, file_path: str) -> bool:
        if callable(self._match):
            return self._match(ctx, file_path)
        patterns = [self._match] if isinstance(self._match, str) else self._match
        for pattern in patterns:
            try:
                if re.fullmatch(pattern, file_path):
                    return True
            except re.error:
                pass
            return fnmatch.fnmatch(file_path, pattern)

    def transform_file(self, ctx: AnyContext, file_path: str):
        if callable(self._transform_file):
            return self._transform_file(ctx, file_path)
        transform_map = {
            keyword: ctx.render(replacement) if self._auto_render else replacement
            for keyword, replacement in self._transform_file.items()
        }
        with open(file_path, "r") as f:
            content = f.read()
        for keyword, replacement in transform_map.items():
            content = content.replace(keyword, replacement)
        with open(file_path, "w") as f:
            f.write(content)



from fastapp_template.common.db_engine import engine
from fastapp_template.config import APP_REPOSITORY_TYPE
from fastapp_template.module.auth.service.user.repository.db_repository import (
    UserDBRepository,
)
from fastapp_template.module.auth.service.user.repository.repository import (
    UserRepository,
)

if APP_REPOSITORY_TYPE == "db":
    user_repository: UserRepository = UserDBRepository(engine)
else:
    user_repository: UserRepository = None



from abc import ABC, abstractmethod

from fastapp_template.schema.user import UserCreate, UserResponse, UserUpdate


class BaseClient(ABC):
    @abstractmethod
    async def get_user_by_id(self, user_id: str) -> UserResponse:
        pass

    @abstractmethod
    async def get_all_users(self) -> list[UserResponse]:
        pass

    @abstractmethod
    async def create_user(
        self, data: UserCreate | list[UserCreate]
    ) -> UserResponse | list[UserResponse]:
        pass

    @abstractmethod
    async def update_user(self, user_id: str, data: UserUpdate) -> UserResponse:
        pass

    @abstractmethod
    async def delete_user(self, user_id: str) -> UserResponse:
        pass



import sys

from zrb.config import INIT_MODULES, INIT_SCRIPTS
from zrb.runner.cli import cli
from zrb.util.cli.style import stylize_error, stylize_warning
from zrb.util.group import NodeNotFoundError
from zrb.util.load import load_file, load_module


def serve_cli():
    try:
        for init_module in INIT_MODULES:
            load_module(init_module)
        for init_script in INIT_SCRIPTS:
            load_file(init_script, -1)
        cli.run(sys.argv[1:])
    except KeyboardInterrupt:
        print(stylize_warning("\nStopped"), file=sys.stderr)
    except NodeNotFoundError as e:
        print(stylize_error(f"{e}"), file=sys.stderr)



import datetime

import ulid
from sqlmodel import Field, SQLModel


class RoleBase(SQLModel):
    name: str


class RoleCreate(RoleBase):
    description: str


class RoleUpdate(SQLModel):
    name: str | None = None
    description: str | None = None


class RoleResponse(RoleBase):
    id: str


class Role(SQLModel, table=True):
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    created_at: datetime.datetime | None
    created_by: str | None
    updated_at: datetime.datetime | None
    updated_by: str | None
    name: str
    description: str



from typing import Any

from zrb.attr.type import StrDictAttr
from zrb.callback.any_callback import AnyCallback
from zrb.session.any_session import AnySession
from zrb.task.any_task import AnyTask
from zrb.util.attr import get_str_dict_attr


class Callback(AnyCallback):

    def __init__(
        self,
        input_mapping: StrDictAttr,
        task: AnyTask,
        auto_render: bool = True,
    ):
        self._input_mapping = input_mapping
        self._task = task
        self._auto_render = auto_render

    async def async_run(self, session: AnySession) -> Any:
        inputs = get_str_dict_attr(
            session.shared_ctx, self._input_mapping, auto_render=self._auto_render
        )
        for name, value in inputs.items():
            session.shared_ctx.input[name] = value
        return await self._task.async_run(session)



import inspect
from collections.abc import Callable
from typing import Any, get_type_hints


def callable_to_tool_schema(
    callable_obj: Callable, name: str | None = None, description: str | None = None
) -> dict[str, Any]:
    """
    Convert a callable into a tool schema dictionary by deriving the parameter schema.

    :param callable_obj: The callable object (e.g., a function).
    :param name: The name to assign to the function in the schema.
    :param description: A description of the function.
    :return: A dictionary representing the tool schema.
    """
    if not callable(callable_obj):
        raise ValueError("Provided object is not callable")
    # Derive name and description
    name = name or callable_obj.__name__
    description = description or (callable_obj.__doc__ or "").strip()
    # Get function signature
    sig = inspect.signature(callable_obj)
    hints = get_type_hints(callable_obj)
    # Build parameter schema
    param_schema = {"type": "object", "properties": {}, "required": []}
    for param_name, param in sig.parameters.items():
        param_type = hints.get(param_name, str)  # Default type is string
        param_schema["properties"][param_name] = {
            "type": _python_type_to_json_type(param_type)
        }
        if param.default is inspect.Parameter.empty:
            param_schema["required"].append(param_name)
    return {
        "type": "function",
        "function": {
            "name": name,
            "description": description,
            "parameters": param_schema,
        },
    }


def _python_type_to_json_type(py_type):
    """
    Map Python types to JSON Schema types.
    """
    if py_type is str:
        return "string"
    elif py_type is int:
        return "integer"
    elif py_type is float:
        return "number"
    elif py_type is bool:
        return "boolean"
    elif py_type is list:
        return "array"
    elif py_type is dict:
        return "object"
    elif py_type in {None, type(None)}:
        return "null"
    else:
        return "string"  # Default to string for unsupported types



import subprocess


def run_shell_command(command: str) -> str:
    """Running a shell command"""
    output = subprocess.check_output(
        command, shell=True, stderr=subprocess.STDOUT, text=True
    )
    return output



import importlib.metadata as metadata
import logging
import os
import platform

from zrb.util.string.conversion import to_boolean


def _get_current_shell() -> str:
    if platform.system() == "Windows":
        return "PowerShell"
    current_shell = os.getenv("SHELL", "")
    if current_shell.endswith("zsh"):
        return "zsh"
    return "bash"


def _get_log_level(level: str) -> int:
    level = level.upper()
    log_levels = {
        "CRITICAL": logging.CRITICAL,  # 50
        "ERROR": logging.ERROR,  # 40
        "WARN": logging.WARNING,  # 30
        "WARNING": logging.WARNING,  # 30
        "INFO": logging.INFO,  # 20
        "DEBUG": logging.DEBUG,  # 10
        "NOTSET": logging.NOTSET,  # 0
    }
    if level in log_levels:
        return log_levels[level]
    return logging.WARNING


DEFAULT_SHELL = os.getenv("ZRB_SHELL", _get_current_shell())
DEFAULT_EDITOR = os.getenv("ZRB_EDITOR", "nano")
INIT_MODULES_STR = os.getenv("ZRB_INIT_MODULES", "")
INIT_MODULES = (
    [module.strip() for module in INIT_MODULES_STR.split(":") if module.strip() != ""]
    if INIT_MODULES_STR != ""
    else []
)
INIT_SCRIPTS_STR = os.getenv("ZRB_INIT_SCRIPTS", "")
INIT_SCRIPTS = (
    [script.strip() for script in INIT_SCRIPTS_STR.split(":") if script.strip() != ""]
    if INIT_SCRIPTS_STR != ""
    else []
)
LOGGING_LEVEL = _get_log_level(os.getenv("ZRB_LOGGING_LEVEL", "WARNING"))
LOAD_BUILTIN = to_boolean(os.getenv("ZRB_LOAD_BUILTIN", "1"))
ENV_PREFIX = os.getenv("ZRB_ENV", "")
SHOW_PROMPT = to_boolean(os.getenv("ZRB_SHOW_PROMPT", "1"))
WARN_UNRECOMMENDED_COMMAND = to_boolean(
    os.getenv("ZRB_WARN_UNRECOMMENDED_COMMAND", "1")
)
SESSION_LOG_DIR = os.getenv(
    "ZRB_SESSION_LOG_DIR", os.path.expanduser(os.path.join("~", ".zrb-session"))
)
TODO_DIR = os.getenv("ZRB_TODO_DIR", os.path.expanduser(os.path.join("~", "todo")))
TODO_VISUAL_FILTER = os.getenv("ZRB_TODO_FILTER", "")
VERSION = metadata.version("zrb")
WEB_HTTP_PORT = int(os.getenv("ZRB_WEB_HTTP_PORT", "21213"))
LLM_MODEL = os.getenv("ZRB_LLM_MODEL", "ollama_chat/llama3.1")
LLM_SYSTEM_PROMPT = os.getenv("ZRB_LLM_SYSTEM_PROMPT", "You are a helpful assistant")
LLM_HISTORY_FILE = os.getenv(
    "ZRB_LLM_HISTORY_FILE",
    os.path.expanduser(os.path.join("~", ".zrb-llm-history.json")),
)
LLM_ALLOW_ACCESS_SHELL = to_boolean(os.getenv("ZRB_LLM_ACCESS_FILE", "1"))
LLM_ALLOW_ACCESS_WEB = to_boolean(os.getenv("ZRB_LLM_ACCESS_WEB", "1"))
RAG_EMBEDDING_MODEL = os.getenv("ZRB_RAG_EMBEDDING_MODEL", "ollama/nomic-embed-text")
RAG_CHUNK_SIZE = int(os.getenv("ZRB_RAG_CHUNK_SIZE", "1024"))
RAG_OVERLAP = int(os.getenv("ZRB_RAG_OVERLAP", "128"))
RAG_MAX_RESULT_COUNT = int(os.getenv("ZRB_RAG_MAX_RESULT_COUNT", "5"))


BANNER = f"""
                bb
   zzzzz rr rr  bb
     zz  rrr  r bbbbbb
    zz   rr     bb   bb
   zzzzz rr     bbbbbb   {VERSION} Janggala
   _ _ . .  . _ .  _ . . .

Your Automation Powerhouse

☕ Donate at: https://stalchmst.com/donation
🐙 Submit issues/PR at: https://github.com/state-alchemists/zrb
🐤 Follow us at: https://twitter.com/zarubastalchmst
"""



from fastapp_template.common.usecase import BaseUsecase
from fastapp_template.module.auth.service.user.repository.factory import user_repository
from fastapp_template.schema.user import UserCreate, UserResponse, UserUpdate


class UserUsecase(BaseUsecase):

    @BaseUsecase.route(
        "/api/v1/users/{user_id}", methods=["get"], response_model=UserResponse
    )
    async def get_user_by_id(self, user_id: str) -> UserResponse:
        return await user_repository.get_by_id(user_id)

    @BaseUsecase.route(
        "/api/v1/users", methods=["get"], response_model=list[UserResponse]
    )
    async def get_all_users(self) -> list[UserResponse]:
        return await user_repository.get_all()

    @BaseUsecase.route(
        "/api/v1/users",
        methods=["post"],
        response_model=UserResponse | list[UserResponse],
    )
    async def create_user(
        self, data: UserCreate | list[UserCreate]
    ) -> UserResponse | list[UserResponse]:
        if isinstance(data, UserCreate):
            return await user_repository.create(data)
        return await user_repository.create_bulk(data)

    @BaseUsecase.route(
        "/api/v1/users/{user_id}", methods=["put"], response_model=UserResponse
    )
    async def update_user(self, user_id: str, data: UserUpdate) -> UserResponse:
        return await user_repository.update(user_id, data)

    @BaseUsecase.route(
        "/api/v1/users/{user_id}", methods=["delete"], response_model=UserResponse
    )
    async def delete_user(self, user_id: str) -> UserResponse:
        return await user_repository.delete(user_id)


user_usecase = UserUsecase()



from collections import deque
from collections.abc import Callable
from typing import Any


class Xcom(deque):

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"<{class_name} {list(self)}>"

    def append(self, value):
        super().append(value)
        self.__call_push_callbacks()

    def push(self, value):
        self.append(value)

    def popleft(self):
        value = super().popleft()
        self.__call_pop_callbacks()
        return value

    def pop(self):
        return self.popleft()

    def popright(self) -> Any:
        value = super().pop()
        self.__call_pop_callbacks()
        return value

    def peek(self):
        if len(self) > 0:
            return self[0]
        else:
            raise IndexError("Xcom is empty")

    def add_push_callback(self, callback: Callable[[], Any]):
        if not hasattr(self, "push_callbacks"):
            self.push_callbacks: list[Callable[[], Any]] = []
        self.push_callbacks.append(callback)

    def add_pop_callback(self, callback: Callable[[], Any]):
        if not hasattr(self, "pop_callbacks"):
            self.pop_callbacks: list[Callable[[], Any]] = []
        self.pop_callbacks.append(callback)

    def __call_push_callbacks(self):
        if not hasattr(self, "push_callbacks"):
            return
        for callback in self.push_callbacks:
            callback()

    def __call_pop_callbacks(self):
        if not hasattr(self, "pop_callbacks"):
            return
        for callback in self.pop_callbacks:
            callback()



import os

from zrb.group.any_group import AnyGroup
from zrb.util.group import get_non_empty_subgroups, get_subtasks
from zrb.util.string.format import fstring_format

_DIR = os.path.dirname(__file__)

with open(os.path.join(_DIR, "view.html"), "r") as f:
    _VIEW_TEMPLATE = f.read()

with open(os.path.join(_DIR, "partial", "group_info.html")) as f:
    _GROUP_INFO_TEMPLATE = f.read()

with open(os.path.join(_DIR, "partial", "group_li.html")) as f:
    _GROUP_LI_TEMPLATE = f.read()

with open(os.path.join(_DIR, "partial", "task_info.html")) as f:
    _TASK_INFO_TEMPLATE = f.read()

with open(os.path.join(_DIR, "partial", "task_li.html")) as f:
    _TASK_LI_TEMPLATE = f.read()


def handle_group_info_ui(root_group: AnyGroup, group: AnyGroup, url: str):
    from fastapi.responses import HTMLResponse

    url_parts = url.split("/")
    parent_url_parts = url_parts[:-2] + [""]
    parent_url = "/".join(parent_url_parts)
    subgroups = get_non_empty_subgroups(group, web_only=True)
    group_info = (
        ""
        if len(subgroups) == 0
        else fstring_format(
            _GROUP_INFO_TEMPLATE,
            {
                "group_li": "\n".join(
                    [
                        fstring_format(
                            _GROUP_LI_TEMPLATE,
                            {
                                "caption": name,
                                "url": url,
                                "description": group.description,
                            },
                        )
                        for name, group in subgroups.items()
                    ]
                )
            },
        )
    )
    subtasks = get_subtasks(group, web_only=True)
    task_info = (
        ""
        if len(subtasks) == 0
        else fstring_format(
            _TASK_INFO_TEMPLATE,
            {
                "task_li": "\n".join(
                    [
                        fstring_format(
                            _TASK_LI_TEMPLATE,
                            {
                                "caption": name,
                                "url": url,
                                "description": task.description,
                            },
                        )
                        for name, task in subtasks.items()
                    ]
                )
            },
        )
    )
    return HTMLResponse(
        fstring_format(
            _VIEW_TEMPLATE,
            {
                "group_info": group_info,
                "task_info": task_info,
                "name": group.name,
                "description": group.description,
                "root_name": root_group.name,
                "root_description": root_group.description,
                "url": url,
                "parent_url": parent_url,
            },
        )
    )



import re

NON_ALPHA_NUM = re.compile(r"[^a-zA-Z0-9]+")
TRUE_STRS = ["true", "1", "yes", "y", "active", "on"]
FALSE_STRS = ["false", "0", "no", "n", "inactive", "off"]


def double_quote(input_string: str) -> str:
    # Escape necessary characters: backslashes and double quotes
    escaped_string = re.sub(r'([\\"])', r"\\\1", input_string)
    # Wrap in double quotes
    return f'"{escaped_string}"'


def to_boolean(text: str) -> bool:
    if text.lower() in TRUE_STRS:
        return True
    if text.lower() in FALSE_STRS:
        return False
    raise Exception(f'Cannot infer boolean value from "{text}"')


def to_camel_case(text: str | None) -> str:
    text = str(text) if text is not None else ""
    pascal = to_pascal_case(text)
    if len(pascal) == 0:
        return pascal
    return pascal[0].lower() + pascal[1:]


def to_pascal_case(text: str | None) -> str:
    text = str(text) if text is not None else ""
    text = _to_alphanum(text)
    return "".join(
        [x.lower().capitalize() for x in _to_space_separated(text).split(" ")]
    )


def to_kebab_case(text: str | None) -> str:
    text = str(text) if text is not None else ""
    text = _to_alphanum(text)
    return "-".join([x.lower() for x in _to_space_separated(text).split(" ")])


def to_snake_case(text: str | None) -> str:
    text = str(text) if text is not None else ""
    text = _to_alphanum(text)
    return "_".join([x.lower() for x in _to_space_separated(text).split(" ")])


def _to_alphanum(text: str | None) -> str:
    text = str(text) if text is not None else ""
    return NON_ALPHA_NUM.sub(" ", text)


def to_human_case(text: str | None) -> str:
    text = str(text) if text is not None else ""
    return " ".join(
        [
            x.lower() if x.upper() != x else x
            for x in _to_space_separated(text).split(" ")
        ]
    )


def _to_space_separated(text: str | None) -> str:
    text = str(text) if text is not None else ""
    text = text.replace("-", " ").replace("_", " ")
    parts = text.split(" ")
    new_parts = []
    for part in parts:
        new_part = ""
        for char_index, char in enumerate(part):
            is_first = char_index == 0
            is_last = char_index == len(part) - 1
            previous_char = part[char_index - 1] if not is_first else ""
            next_char = part[char_index + 1] if not is_last else ""
            if (
                char.isupper()
                and char != " "
                and (
                    (not is_last and next_char.islower())
                    or (not is_first and previous_char.islower())
                )
            ):
                new_part += " " + char
                continue
            new_part += char
        new_part = new_part.strip(" ")
        if new_part != "":
            new_parts.append(new_part)
    return " ".join(new_parts).strip(" ")



import os

from fastapp_template._zrb.config import ACTIVATE_VENV_SCRIPT, APP_DIR
from fastapp_template._zrb.group import (
    app_create_group,
    app_migrate_group,
    app_run_group,
)
from fastapp_template._zrb.helper import (
    create_migration,
    migrate_module,
    run_microservice,
)
from fastapp_template._zrb.venv_task import prepare_venv

from zrb import CmdTask, Env, EnvFile, Task

# 🚀 Run/Migrate All ===========================================================

run_all = app_run_group.add_task(
    Task(
        name="run-app-name", description="🟢 Run App Name as monolith and microservices"
    ),
    alias="all",
)

migrate_all = app_migrate_group.add_task(
    Task(
        name="migrate-app-name",
        description="📦 Run App Name DB migration for monolith and microservices",
    ),
    alias="all",
)

create_all_migration = app_create_group.add_task(
    Task(
        name="create-app-name-migration", description="📦 Create App Name DB migration"
    ),
    alias="migration",
)

# 🗿 Run/Migrate Monolith =====================================================

run_monolith = app_run_group.add_task(
    CmdTask(
        name="run-monolith-app-name",
        description="🗿 Run App Name as a monolith",
        env=[
            EnvFile(path=os.path.join(APP_DIR, "template.env")),
            Env(name="APP_NAME_MODE", default="monolith"),
        ],
        cwd=APP_DIR,
        cmd=[
            ACTIVATE_VENV_SCRIPT,
            'fastapi dev main.py --port "${APP_NAME_PORT}"',
        ],
        render_cmd=False,
        retries=2,
    ),
    alias="monolith",
)
prepare_venv >> run_monolith >> run_all

migrate_monolith = app_migrate_group.add_task(
    Task(
        name="migrate-monolith-app-name",
        description="🗿 Run App Name DB migration for monolith",
    ),
    alias="monolith",
)
migrate_monolith >> migrate_all

# 🌐 Run/Migrate Microsevices ==================================================

run_microservices = app_run_group.add_task(
    Task(
        name="run-microservices-app-name",
        description="🌐 Run App Name as microservices",
    ),
    alias="microservices",
)
run_microservices >> run_all

migrate_microservices = app_migrate_group.add_task(
    Task(
        name="migrate-microservices-app-name",
        description="🌐 Run App Name DB migration for microservices",
    ),
    alias="microservices",
)
migrate_microservices >> migrate_all

# 📡 Run/Migrate Gateway =======================================================

run_gateway = app_run_group.add_task(
    run_microservice("gateway", 3001, "gateway"), alias="microservices-gateway"
)
prepare_venv >> run_gateway >> run_microservices

create_gateway_migration = app_create_group.add_task(
    create_migration("gateway", "gateway"), alias="gateway-migration"
)
prepare_venv >> create_gateway_migration >> create_all_migration

migrate_monolith_gateway = migrate_module("gateway", "gateway", as_microservices=False)
prepare_venv >> migrate_monolith_gateway >> [migrate_monolith, run_monolith]

migrate_microservices_gateway = app_migrate_group.add_task(
    migrate_module("gateway", "gateway", as_microservices=True),
    alias="microservices-gateway",
)
prepare_venv >> migrate_microservices_gateway >> [migrate_microservices, run_gateway]

# 🔐 Run/Migrate Auth ==========================================================

run_auth = app_run_group.add_task(
    run_microservice("auth", 3002, "auth"), alias="microservices-auth"
)
prepare_venv >> run_auth >> run_microservices

create_auth_migration = app_create_group.add_task(
    create_migration("auth", "auth"), alias="auth-migration"
)
prepare_venv >> create_auth_migration >> create_all_migration

migrate_monolith_auth = migrate_module("auth", "auth", as_microservices=False)
prepare_venv >> migrate_monolith_auth >> [migrate_monolith, run_monolith]

migrate_microservices_auth = app_migrate_group.add_task(
    migrate_module("auth", "auth", as_microservices=True), alias="microservices-auth"
)
prepare_venv >> migrate_microservices_auth >> [migrate_microservices, run_auth]



import os

from zrb.builtin.group import add_to_project_group
from zrb.context.any_context import AnyContext
from zrb.input.str_input import StrInput
from zrb.task.make_task import make_task
from zrb.task.scaffolder import Scaffolder
from zrb.task.task import Task
from zrb.util.string.conversion import double_quote
from zrb.util.string.name import get_random_name

_DIR = os.path.dirname(__file__)


scaffold_fastapp = Scaffolder(
    name="scaffold-fastapp",
    input=[
        StrInput(
            name="project-dir",
            description="Project directory",
            prompt="Project directory",
            default_str=lambda _: os.getcwd(),
        ),
        StrInput(
            name="app-name",
            description="App name",
            prompt="App name",
            default_str=lambda _: get_random_name(separator="_"),
        ),
    ],
    source_path=os.path.join(_DIR, "fastapp_template"),
    render_source_path=False,
    destination_path=lambda ctx: os.path.join(
        ctx.input["project-dir"], ctx.input["app-name"]
    ),
    transform_content={
        "fastapp_template": "{to_snake_case(ctx.input['app-name'])}",
        "App Name": "{ctx.input['app-name'].title()}",
        "App name": "{ctx.input['app-name'].capitalize()}",
        "app-name": "{to_kebab_case(ctx.input['app-name'])}",
        "app_name": "{to_snake_case(ctx.input['app-name'])}",
        "APP_NAME": "{to_snake_case(ctx.input['app-name']).upper()}",
        "secure-password": lambda _: get_random_name(),
    },
    retries=0,
)


@make_task(
    name="register-fastapp-automation",
)
def register_fastapp_automation(ctx: AnyContext):
    project_dir_path = ctx.input["project-dir"]
    zrb_init_path = os.path.join(project_dir_path, "zrb_init.py")
    app_dir_path = ctx.input["app-name"]
    app_automation_file_part = ", ".join(
        [double_quote(part) for part in [app_dir_path, "_zrb", "main.py"]]
    )
    with open(zrb_init_path, "+a") as f:
        f.write(f"load_file(os.path.join(_DIR, {app_automation_file_part}))\n")


scaffold_fastapp >> register_fastapp_automation

add_fastapp_to_project = add_to_project_group.add_task(
    Task(
        name="add-fastapp",
        description="🚀 Add FastApp to project",
    ),
    alias="fastapp",
)
add_fastapp_to_project << [register_fastapp_automation]



import libcst as cst


class ClassPropertyAdder(cst.CSTTransformer):
    def __init__(
        self, class_name: str, property_name: str, annotation: str, default_value: str
    ):
        self.class_name = class_name
        self.property_name = property_name
        self.annotation = cst.Annotation(cst.parse_expression(annotation))
        self.default_value = cst.parse_expression(default_value)
        self.class_found = False

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        # Check if this is the target class
        if original_node.name.value == self.class_name:
            self.class_found = True
            # Create the annotated property with a default value
            new_property = cst.SimpleStatementLine(
                body=[
                    cst.AnnAssign(
                        target=cst.Name(self.property_name),
                        annotation=self.annotation,
                        value=self.default_value,
                    )
                ]
            )
            # Insert the new property at the start of the class body with a newline
            new_body = cst.IndentedBlock(body=(new_property,) + updated_node.body.body)
            return updated_node.with_changes(body=new_body)
        return updated_node


def add_property_to_class(
    original_code: str,
    class_name: str,
    property_name: str,
    annotation: str,
    default_value: str,
) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize transformer with the class name, property name, annotation, and default value
    transformer = ClassPropertyAdder(
        class_name, property_name, annotation, default_value
    )
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Check if the class was found
    if not transformer.class_found:
        raise ValueError(f"Class {class_name} not found in the provided code.")
    # Return the modified code
    return modified_module.code



from abc import ABC, abstractmethod

from fastapp_template.schema.user import User, UserCreate, UserResponse, UserUpdate


class UserRepository(ABC):

    @abstractmethod
    async def create(self, user_data: UserCreate) -> UserResponse:
        pass

    @abstractmethod
    async def get_by_id(self, user_id: str) -> User:
        pass

    @abstractmethod
    async def get_all(self) -> list[User]:
        pass

    @abstractmethod
    async def update(self, user_id: str, user_data: UserUpdate) -> User:
        pass

    @abstractmethod
    async def delete(self, user_id: str) -> User:
        pass

    @abstractmethod
    async def create_bulk(self, user_data_list: list[UserCreate]) -> list[UserResponse]:
        pass

    @abstractmethod
    async def get_by_credentials(self, username: str, password: str) -> UserResponse:
        pass



import libcst as cst


class ParentClassAdder(cst.CSTTransformer):
    def __init__(self, class_name: str, parent_class_name: str):
        self.class_name = class_name
        self.parent_class_name = parent_class_name
        self.class_found = False

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        # Check if this is the target class
        if original_node.name.value == self.class_name:
            self.class_found = True
            # Add the parent class to the existing bases
            new_bases = (
                cst.Arg(value=cst.Name(self.parent_class_name)),
                *updated_node.bases,
            )
            return updated_node.with_changes(bases=new_bases)
        return updated_node


def add_parent_to_class(
    original_code: str, class_name: str, parent_class_name: str
) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize transformer with the class name and parent class name
    transformer = ParentClassAdder(class_name, parent_class_name)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Check if the class was found
    if not transformer.class_found:
        raise ValueError(f"Class {class_name} not found in the provided code.")
    # Return the modified code
    return modified_module.code



from enum import Enum
from functools import partial, wraps
from typing import Any, Callable, Sequence

import httpx
from fastapi import APIRouter, params
from fastapi.routing import APIRoute
from fastapi.types import IncEx
from fastapi.utils import generate_unique_id
from starlette.responses import JSONResponse, Response


class RouteParam:
    def __init__(
        self,
        path: str,
        response_model: Any,
        status_code: int | None = None,
        tags: list[str | Enum] | None = None,
        dependencies: Sequence[params.Depends] | None = None,
        summary: str | None = None,
        description: str = "",
        response_description: str = "",
        responses: dict[int | str, dict[str, Any]] | None = None,
        deprecated: bool | None = None,
        methods: set[str] | list[str] | None = None,
        operation_id: str | None = None,
        response_model_include: IncEx | None = None,
        response_model_exclude: IncEx | None = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: type[Response] = Response,
        name: str | None = None,
        openapi_extra: dict[str, Any] | None = None,
        generate_unique_id_function: Callable[[APIRoute], str] | None = None,
        func: Callable | None = None,
    ):
        self.path = path
        self.response_model = response_model
        self.status_code = status_code
        self.tags = tags
        self.dependencies = dependencies
        self.summary = summary
        self.description = description
        self.response_description = response_description
        self.responses = responses
        self.deprecated = deprecated
        self.methods = methods
        self.operation_id = operation_id
        self.response_model_include = response_model_include
        self.response_model_exclude = response_model_exclude
        self.response_model_by_alias = response_model_by_alias
        self.response_model_exclude_unset = response_model_exclude_unset
        self.response_model_exclude_defaults = response_model_exclude_defaults
        self.response_model_exclude_none = response_model_exclude_none
        self.include_in_schema = include_in_schema
        self.response_class = response_class
        self.name = name
        self.openapi_extra = openapi_extra
        self.generate_unique_id_function = generate_unique_id_function
        self.func = func


class BaseUsecase:
    _route_params: dict[str, RouteParam] = {}

    @classmethod
    def route(
        cls,
        path: str,
        *,
        response_model: Any = None,
        status_code: int | None = None,
        tags: list[str | Enum] | None = None,
        dependencies: Sequence[params.Depends] | None = None,
        summary: str | None = None,
        description: str = None,
        response_description: str = "Successful Response",
        responses: dict[int | str, dict[str, Any]] | None = None,
        deprecated: bool | None = None,
        methods: set[str] | list[str] | None = None,
        operation_id: str | None = None,
        response_model_include: IncEx | None = None,
        response_model_exclude: IncEx | None = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: type[Response] = JSONResponse,
        name: str | None = None,
        openapi_extra: dict[str, Any] | None = None,
        generate_unique_id_function: Callable[[APIRoute], str] = generate_unique_id,
    ):
        """
        Decorator to register a method with its HTTP details.
        """

        def decorator(func: Callable):
            cls._route_params[func.__name__] = RouteParam(
                path=path,
                response_model=response_model,
                status_code=status_code,
                tags=tags,
                dependencies=dependencies,
                summary=summary,
                description=description,
                response_description=response_description,
                responses=responses,
                deprecated=deprecated,
                methods=methods,
                operation_id=operation_id,
                response_model_include=response_model_include,
                response_model_exclude=response_model_exclude,
                response_model_by_alias=response_model_by_alias,
                response_model_exclude_unset=response_model_exclude_unset,
                response_model_exclude_defaults=response_model_exclude_defaults,
                response_model_exclude_none=response_model_exclude_none,
                include_in_schema=include_in_schema,
                response_class=response_class,
                name=name,
                openapi_extra=openapi_extra,
                generate_unique_id_function=generate_unique_id_function,
                func=func,
            )

            @wraps(func)
            async def wrapped(*args, **kwargs):
                return await func(*args, **kwargs)

            return wrapped

        return decorator

    def as_direct_client(self):
        """
        Dynamically create a direct client class.
        """
        _methods = self._route_params
        DirectClient = create_client_class("DirectClient")
        for name, details in _methods.items():
            func = details.func
            client_method = create_direct_client_method(func, self)
            setattr(DirectClient, name, client_method.__get__(DirectClient))
        return DirectClient

    def as_api_client(self, base_url: str):
        """
        Dynamically create an API client class.
        """
        _methods = self._route_params
        APIClient = create_client_class("APIClient")
        # Dynamically generate methods
        for name, param in _methods.items():
            client_method = create_api_client_method(param, base_url)
            setattr(APIClient, name, client_method.__get__(APIClient))
        return APIClient

    def serve_route(self, app: APIRouter):
        """
        Dynamically add routes to FastAPI.
        """
        for _, route_param in self._route_params.items():
            bound_func = partial(route_param.func, self)
            bound_func.__name__ = route_param.func.__name__
            bound_func.__doc__ = route_param.func.__doc__
            app.add_api_route(
                path=route_param.path,
                endpoint=bound_func,
                response_model=route_param.response_model,
                status_code=route_param.status_code,
                tags=route_param.tags,
                dependencies=route_param.dependencies,
                summary=route_param.summary,
                description=route_param.description,
                response_description=route_param.response_description,
                responses=route_param.responses,
                deprecated=route_param.deprecated,
                methods=route_param.methods,
                operation_id=route_param.operation_id,
                response_model_include=route_param.response_model_include,
                response_model_exclude=route_param.response_model_exclude,
                response_model_by_alias=route_param.response_model_by_alias,
                response_model_exclude_unset=route_param.response_model_exclude_unset,
                response_model_exclude_defaults=route_param.response_model_exclude_defaults,
                response_model_exclude_none=route_param.response_model_exclude_none,
                include_in_schema=route_param.include_in_schema,
                response_class=route_param.response_class,
                name=route_param.name,
                openapi_extra=route_param.openapi_extra,
                generate_unique_id_function=route_param.generate_unique_id_function,
            )


def create_client_class(name):
    class Client:
        pass

    Client.__name__ = name
    return Client


def create_direct_client_method(func: Callable, usecase: BaseUsecase):
    async def client_method(self, *args, **kwargs):
        return await func(usecase, *args, **kwargs)

    return client_method


def create_api_client_method(param: RouteParam, base_url: str):
    _url = param.path
    _methods = [method.lower() for method in param.methods]

    async def client_method(self, *args, **kwargs):
        async with httpx.AsyncClient() as client:
            url = base_url + _url.format(**kwargs)
            if "post" in _methods:
                response = await client.post(url, json=kwargs)
            elif "put" in _methods:
                response = await client.put(url, json=kwargs)
            elif "delete" in _methods:
                response = await client.delete(url, json=kwargs)
            else:
                response = await client.get(url, params=kwargs)
            # Add more HTTP methods as needed
            response.raise_for_status()
            return response.json()

    return client_method



from fastapp_template.common.app import app
from fastapp_template.module.auth import route as auth_route
from fastapp_template.module.gateway import route as gateway_route

assert app
assert gateway_route
assert auth_route



import datetime

import ulid
from sqlmodel import Field, SQLModel


class UserBase(SQLModel):
    username: str


class UserCreate(UserBase):
    password: str


class UserUpdate(SQLModel):
    username: str | None = None
    password: str | None = None


class UserResponse(UserBase):
    id: str


class User(SQLModel, table=True):
    id: str = Field(default_factory=lambda: ulid.new().str, primary_key=True)
    created_at: datetime.datetime
    created_by: str
    updated_at: datetime.datetime
    updated_by: str
    username: str
    password: str



from abc import ABC, abstractmethod
from collections.abc import Callable

from zrb.attr.type import fstring
from zrb.context.context import Context


class AnyCmdVal(ABC):
    @abstractmethod
    def to_str(self, ctx: Context) -> str:
        pass


class CmdPath(AnyCmdVal):
    def __init__(self, path: str, auto_render: bool = True):
        self._path = path
        self._auto_render = auto_render

    def to_str(self, ctx: Context) -> str:
        file_path = ctx.render(self._path) if self._auto_render else self._path
        with open(file_path) as file:
            return file.read()


class Cmd(AnyCmdVal):
    def __init__(self, cmd: str, auto_render: str):
        self._cmd = cmd
        self._auto_render = auto_render

    def to_str(self, ctx: Context) -> str:
        return ctx.render(self._cmd) if self._auto_render else self._cmd


SingleCmdVal = AnyCmdVal | fstring | Callable[[Context], str]
CmdVal = SingleCmdVal | list[SingleCmdVal]



from zrb.builtin.group import python_group
from zrb.task.cmd_task import CmdTask

format_python_code = python_group.add_task(
    CmdTask(
        name="format-code",
        description="✏️ Format Python code",
        cmd=["isort .", "black ."],
    ),
    alias="format",
)



import libcst as cst


class FunctionCallParamAdder(cst.CSTTransformer):
    def __init__(self, func_name: str, new_param: str):
        self.func_name = func_name
        # Parse the new parameter to ensure it’s a valid CST node
        self.new_param = cst.parse_expression(new_param)
        self.param_added = False

    def leave_Call(self, original_node: cst.Call, updated_node: cst.Call) -> cst.Call:
        # Check if the function call name matches the target function
        if (
            isinstance(original_node.func, cst.Name)
            and original_node.func.value == self.func_name
        ):  # noqa
            # Add the new parameter to the function call arguments
            new_args = updated_node.args + (cst.Arg(value=self.new_param),)
            self.param_added = True
            return updated_node.with_changes(args=new_args)
        return updated_node


def add_param_to_function_call(
    original_code: str, func_name: str, new_param: str
) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize the transformer with the necessary information
    transformer = FunctionCallParamAdder(func_name, new_param)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Error handling: raise an error if the function call is not found
    if not transformer.param_added:
        raise ValueError(
            f"Function call to {func_name} not found in the provided code."
        )
    # Return the modified code
    return modified_module.code



from __future__ import annotations  # Enables forward references

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput

if TYPE_CHECKING:
    from zrb.context import any_context
    from zrb.session import session


class AnyTask(ABC):
    """Abstract base class representing a task in the system.

    This class defines the interface for tasks, including methods for retrieving
    task metadata (such as name, color, and description), handling inputs and
    environment variables, and managing task dependencies (upstreams and fallbacks).
    It also includes methods for running tasks synchronously or asynchronously.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Task name"""
        pass

    @property
    @abstractmethod
    def color(self) -> int | None:
        """Task color, if any."""
        pass

    @property
    @abstractmethod
    def icon(self) -> str | None:
        """Task icon, if any."""
        pass

    @property
    @abstractmethod
    def description(self) -> str:
        """Task description"""
        pass

    @property
    @abstractmethod
    def cli_only(self) -> bool:
        """Whether the task is CLI only or not"""
        pass

    @property
    @abstractmethod
    def inputs(self) -> list[AnyInput]:
        """Task inputs"""
        pass

    @property
    @abstractmethod
    def envs(self) -> list[AnyEnv]:
        """Task envs"""
        pass

    @property
    @abstractmethod
    def upstreams(self) -> list["AnyTask"]:
        """Task upstreams"""
        pass

    @property
    @abstractmethod
    def fallbacks(self) -> list["AnyTask"]:
        """Task fallbacks"""
        pass

    @property
    @abstractmethod
    def successors(self) -> list["AnyTask"]:
        """Task successors"""
        pass

    @property
    @abstractmethod
    def readiness_checks(self) -> list["AnyTask"]:
        """Task readiness checks"""
        pass

    @abstractmethod
    def append_fallback(self, fallbacks: "AnyTask" | list["AnyTask"]):
        """Add the fallback tasks.

        Args:
            fallbacks (AnyTask | list[AnyTask]): A single fallback task or
                a list of fallback tasks.
        """
        pass

    @abstractmethod
    def append_successor(self, successors: "AnyTask" | list["AnyTask"]):
        """Add the successor tasks.

        Args:
            successors (AnyTask | list[AnyTask]): A single successor task or
                a list of successor tasks.
        """
        pass

    @abstractmethod
    def append_readiness_check(self, readiness_checks: "AnyTask" | list["AnyTask"]):
        """Add the readiness_check tasks.

        Args:
            readiness_checks (AnyTask | list[AnyTask]): A single readiness_check task or
                a list of readiness_check tasks.
        """
        pass

    @abstractmethod
    def append_upstream(self, upstreams: "AnyTask" | list["AnyTask"]):
        """Add the upstream tasks that this task depends on.

        Args:
            upstreams (AnyTask | list[AnyTask]): A single upstream task or
                a list of upstream tasks.
        """
        pass

    @abstractmethod
    def get_ctx(self, session: session.AnySession) -> any_context.AnyContext:
        pass

    @abstractmethod
    def run(
        self, session: session.AnySession | None = None, str_kwargs: dict[str, str] = {}
    ) -> Any:
        """Runs the task synchronously.

        Args:
            session (AnySession): The shared session.
            str_kwargs(dict[str, str]): The input string values.

        Returns:
            Any: The result of the task execution.
        """
        pass

    @abstractmethod
    async def async_run(
        self, session: session.AnySession | None = None, str_kwargs: dict[str, str] = {}
    ) -> Any:
        """Runs the task asynchronously.

        Args:
            session (AnySession): The shared session.
            str_kwargs(dict[str, str]): The input string values.

        Returns:
            Any: The result of the task execution.
        """
        pass

    @abstractmethod
    async def exec_root_tasks(self, session: session.AnySession):
        """Execute the root tasks along with the downstreams until the current task
        is ready.

        Args:
            session (AnySession): The shared session.
        """
        pass

    @abstractmethod
    async def exec_chain(self, session: session.AnySession):
        """Execute the task along with the downstreams.

        Args:
            session (AnySession): The shared session.
        """
        pass

    @abstractmethod
    async def exec(self, session: session.AnySession):
        """Execute the task (without upstream or downstream).

        Args:
            session (AnySession): The shared session.
        """
        pass



import os
import subprocess

from pydantic import BaseModel


class DiffResult(BaseModel):
    created: list[str]
    removed: list[str]
    updated: list[str]


def get_diff(repo_dir: str, source_commit: str, current_commit: str) -> DiffResult:
    try:
        result = subprocess.run(
            ["git", "diff", source_commit, current_commit],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)
    lines = result.stdout.strip().split("\n")
    diff: dict[str, dict[str, bool]] = {}
    for line in lines:
        if not line.startswith("---") and not line.startswith("+++"):
            continue
        if line[4:6] != "a/" and line[4:6] != "b/":
            continue
        # line should contains something like `--- a/some-file.txt`
        file = line[6:]
        if file not in diff:
            diff[file] = {"plus": False, "minus": False}
        if line.startswith("---"):
            diff[file]["minus"] = True
        if line.startswith("+++"):
            diff[file]["plus"] = True
    return DiffResult(
        created=[
            file for file, state in diff.items() if state["plus"] and not state["minus"]
        ],
        removed=[
            file for file, state in diff.items() if not state["plus"] and state["minus"]
        ],
        updated=[
            file for file, state in diff.items() if state["plus"] and state["minus"]
        ],
    )


def get_repo_dir() -> str:
    try:
        # Run the Git command to get the repository's top-level directory
        result = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            check=True,
        )
        # Return the directory path
        return os.path.abspath(result.stdout.strip())
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def get_current_branch(repo_dir: str) -> str:
    try:
        result = subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def get_branches(repo_dir: str) -> list[str]:
    try:
        result = subprocess.run(
            ["git", "branch"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
        return [
            branch.lstrip("*").strip() for branch in result.stdout.strip().split("\n")
        ]
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def delete_branch(repo_dir: str, branch_name: str) -> str:
    try:
        result = subprocess.run(
            ["git", "branch", "-D", branch_name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def add(repo_dir: str) -> str:
    try:
        subprocess.run(
            ["git", "add", ".", "-A"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def commit(repo_dir: str, message: str) -> str:
    try:
        subprocess.run(
            ["git", "commit", "-m", message],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        ignored_error_message = "nothing to commit, working tree clean"
        if (
            ignored_error_message not in e.stderr
            and ignored_error_message not in e.stdout
        ):
            raise Exception(e.stderr or e.stdout)


def pull(repo_dir: str, remote: str, branch: str) -> str:
    try:
        subprocess.run(
            ["git", "pull", remote, branch],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def push(repo_dir: str, remote: str, branch: str) -> str:
    try:
        subprocess.run(
            ["git", "push", "-u", remote, branch],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)



import json
from typing import Annotated


def open_web_page(url: str) -> str:
    """Get content from a web page."""
    import requests

    response = requests.get(
        url,
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"  # noqa
        },
    )
    if response.status_code != 200:
        raise Exception(
            f"Error: Unable to retrieve search results (status code: {response.status_code})"  # noqa
        )
    return json.dumps(parse_html_text(response.text))


def query_internet(
    query: Annotated[str, "Search query"],
    num_results: Annotated[int, "Search result count, by default 10"] = 10,
) -> str:
    """Search factual information from the internet by using Google."""
    import requests

    response = requests.get(
        "https://google.com/search",
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36"  # noqa
        },
        params={
            "q": query,
            "num": num_results,
            "hl": "en",
            "safe": "off",
        },
    )
    if response.status_code != 200:
        raise Exception(
            f"Error: Unable to retrieve search results (status code: {response.status_code})"  # noqa
        )
    return json.dumps(parse_html_text(response.text))


def parse_html_text(html_text: str) -> dict[str, str]:
    from bs4 import BeautifulSoup

    ignored_tags = [
        "script",
        "link",
        "meta",
        "style",
        "code",
        "footer",
        "nav",
        "header",
        "aside",
    ]
    soup = BeautifulSoup(html_text, "html.parser")
    links = []
    for anchor in soup.find_all("a"):
        if not anchor or "href" not in anchor.attrs:
            continue
        link: str = anchor["href"]
        if link.startswith("#") or link.startswith("/"):
            continue
        links.append(link)
    for tag in soup(ignored_tags):
        tag.decompose()
    html_text = soup.get_text(separator=" ", strip=True)
    return {"content": html_text, "links_on_page": links}



from fastapp_template.module.auth.client.base_client import BaseClient
from fastapp_template.module.auth.service.user.usecase import user_usecase


class DirectClient(user_usecase.as_direct_client(), BaseClient):
    pass



import os

from zrb.context.any_context import AnyContext


def get_install_prerequisites_cmd(ctx: AnyContext) -> str:
    package_manager: str = ctx.input["package-manager"]
    if package_manager in ["brew", "spack"]:
        cmd = f"{package_manager} install coreutils curl git"
    elif package_manager == "pacman":
        cmd = f"{package_manager} -S curl git"
    else:
        cmd = f"{package_manager} install curl git"
    use_sudo: bool = ctx.input["use-sudo"]
    if use_sudo:
        return f"sudo {cmd}"
    return cmd


def check_inexist_asdf_dir(_: AnyContext):
    asdf_dir = os.path.expanduser(os.path.join("~", ".asdf"))
    return not os.path.isdir(asdf_dir)


def setup_asdf_sh_config(file_path: str):
    _setup_asdf_config(file_path, '. "$HOME/.asdf/asdf.sh"')


def setup_asdf_ps_config(file_path: str):
    _setup_asdf_config(file_path, '. "$HOME/.asdf/asdf.ps1"')


def _setup_asdf_config(file_path: str, asdf_config: str):
    dir_path = os.path.dirname(file_path)
    os.makedirs(dir_path, exist_ok=True)
    if not os.path.isfile(file_path):
        with open(file_path, "w") as f:
            f.write("")
    with open(file_path, "r") as f:
        content = f.read()
    if asdf_config in content:
        return
    with open(file_path, "a") as f:
        f.write(f"\n{asdf_config}\n")



import os
import platform

DIR = os.path.dirname(__file__)
APP_DIR = os.path.dirname(DIR)
APP_MODULE_NAME = os.path.basename(APP_DIR)

MICROSERVICES_ENV_VARS = {
    "APP_NAME_MODE": "microservices",
    "APP_NAME_AUTH_BASE_URL": "http://localhost:3002",
}
MONOLITH_ENV_VARS = {"APP_NAME_MODE": "monolith"}

if platform.system() == "Windows":
    ACTIVATE_VENV_SCRIPT = "Set-ExecutionPolicy -ExecutionPolicy RemoteSigned -Scope CurrentUser; . .venv\Scripts\Activate"  # noqa
else:
    ACTIVATE_VENV_SCRIPT = "source .venv/bin/activate"



import os

APP_PATH = os.path.dirname(__file__)

APP_MODE = os.getenv("APP_NAME_MODE", "monolith")
APP_MODULES = [
    module.strip()
    for module in os.getenv("APP_NAME_MODULES", "").split(",")
    if module.strip() != ""
]
APP_PORT = int(os.getenv("APP_NAME_PORT", "3000"))
APP_COMMUNICATION = os.getenv(
    "APP_NAME_COMMUNICATION", "direct" if APP_MODE == "monolith" else "api"
)
APP_REPOSITORY_TYPE = os.getenv("APP_REPOSITORY_TYPE", "db")
APP_DB_URL = os.getenv(
    "APP_DB_URL",
    (
        f"sqlite:///{APP_PATH}/monolith.db"
        if APP_MODE == "monolith" or len(APP_MODULES) == 0
        else f"sqlite:///{APP_PATH}/{APP_MODULES[0]}_microservices.db"
    ),
)
APP_AUTH_SUPER_USER = os.getenv("APP_NAME_AUTH_SUPER_USER", "admin")
APP_AUTH_SUPER_USER_PASSWORD = os.getenv(
    "APP_NAME_AUTH_SUPER_USER_PASSWORD", "secure-password"
)

APP_AUTH_BASE_URL = os.getenv("APP_NAME_AUTH_BASE_URL", "http://localhost:3001")



import libcst as cst


class FunctionCodeAdder(cst.CSTTransformer):
    def __init__(self, function_name: str, new_code: str):
        self.function_name = function_name
        # Use parse_module to handle multiple statements
        self.new_code = cst.parse_statement(new_code)
        self.function_found = False

    def leave_FunctionDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        # Check if the class matches the target class
        if original_node.name.value == self.function_name:
            self.function_found = True
            # Add the method to the class body
            new_body = updated_node.body.with_changes(
                body=updated_node.body.body + (self.new_code,)
            )
            return updated_node.with_changes(body=new_body)
        return updated_node


def add_code_to_function(original_code: str, function_name: str, new_code: str) -> str:
    # Parse the original code into a module
    module = cst.parse_module(original_code)
    # Initialize the transformer with the necessary information
    transformer = FunctionCodeAdder(function_name, new_code)
    # Apply the transformation
    modified_module = module.visit(transformer)
    # Error handling: raise an error if the class or function is not found
    if not transformer.function_found:
        raise ValueError(f"Function {function_name} not found.")
    # Return the modified code
    return modified_module.code



from fastapp_template.common.app import app
from fastapp_template.common.schema import BasicResponse
from fastapp_template.config import APP_MODE, APP_MODULES
from fastapp_template.module.auth.client.factory import client as auth_client
from fastapp_template.schema.user import UserCreate, UserResponse

if APP_MODE == "monolith" or "gateway" in APP_MODULES:

    if APP_MODE == "monolith" or (len(APP_MODULES) > 0 and APP_MODULES[0] == "gateway"):

        @app.api_route("/health", methods=["GET", "HEAD"], response_model=BasicResponse)
        async def health():
            return BasicResponse(message="ok")

        @app.api_route(
            "/readiness", methods=["GET", "HEAD"], response_model=BasicResponse
        )
        async def readiness():
            return BasicResponse(message="ok")

    @app.get("/api/v1/users", response_model=list[UserResponse])
    async def auth_get_all_users() -> UserResponse:
        return await auth_client.get_all_users()

    @app.post("/api/v1/users", response_model=UserResponse | list[UserResponse])
    async def auth_create_user(data: UserCreate | list[UserCreate]):
        return await auth_client.create_user(data)



import asyncio
from collections.abc import Callable

from zrb.attr.type import StrAttr
from zrb.context.any_context import AnyContext
from zrb.context.context import Context
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_str_attr


class HttpCheck(BaseTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput] | AnyInput | None = None,
        env: list[AnyEnv] | AnyEnv | None = None,
        url: StrAttr = "http://localhost",
        render_url: bool = True,
        http_method: StrAttr = "GET",
        interval: int = 5,
        execute_condition: bool | str | Callable[[Context], bool] = True,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=0,
            upstream=upstream,
            fallback=fallback,
        )
        self._url = url
        self._render_url = render_url
        self._http_method = http_method
        self._interval = interval

    def _get_url(self, ctx: AnyContext) -> str:
        return get_str_attr(
            ctx, self._url, "http://localhost", auto_render=self._render_url
        )

    def _get_http_method(self, ctx: AnyContext) -> str:
        return get_str_attr(ctx, self._http_method, "GET", auto_render=True).upper()

    async def _exec_action(self, ctx: AnyContext) -> bool:
        import requests

        url = self._get_url(ctx)
        http_method = self._get_http_method(ctx)
        while True:
            try:
                response = requests.request(http_method, url)
                if response.status_code == 200:
                    return response
                ctx.log_info(f"HTTP Status code: {response.status_code}")
            except asyncio.TimeoutError as e:
                ctx.log_info(f"Timeout error {e}")
            except Exception as e:
                ctx.log_info(f"Error: {e}")
            await asyncio.sleep(self._interval)



import asyncio
import inspect
from typing import Any


async def run_async(value: Any) -> Any:
    if inspect.isawaitable(value):
        return await value
    return await asyncio.to_thread(lambda: value)



from zrb import Group
from zrb.builtin import project_group

app_group = project_group.add_group(
    Group(name="app-name", description="🚀 Managing App Name")
)

app_run_group = app_group.add_group(Group(name="run", description="🟢 Run App Name"))

app_migrate_group = app_group.add_group(
    Group(name="migrate", description="📦 Run App Name DB migration")
)

app_create_group = app_group.add_group(
    Group(name="create", description="✨ Create resources for App Name")
)



from typing import Any

from zrb.attr.type import StrAttr
from zrb.context.any_shared_context import AnySharedContext
from zrb.input.any_input import AnyInput
from zrb.util.attr import get_str_attr


class BaseInput(AnyInput):
    def __init__(
        self,
        name: str,
        description: str | None = None,
        prompt: str | None = None,
        default_str: StrAttr = "",
        auto_render: bool = True,
        allow_empty: bool = False,
    ):
        self._name = name
        self._description = description
        self._prompt = prompt
        self._default_str = default_str
        self._auto_render = auto_render
        self._allow_empty = allow_empty

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self._name}>"

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description if self._description is not None else self.name

    @property
    def prompt_message(self) -> str:
        return self._prompt if self._prompt is not None else self.name

    def to_html(self, ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = self._get_default_str(ctx)
        return f'<input name="{name}" placeholder="{description}" value="{default}" />'

    def update_shared_context(
        self, shared_ctx: AnySharedContext, str_value: str | None = None
    ):
        if str_value is None:
            str_value = self._get_default_str(shared_ctx)
        shared_ctx.input[self.name] = self._parse_str_value(str_value)

    def _parse_str_value(self, str_value: str) -> Any:
        """Override this to transform str_value"""
        return str_value

    def prompt_cli_str(self, shared_ctx: AnySharedContext) -> str:
        value = self._prompt_cli_str(shared_ctx)
        while not self._allow_empty and value == "":
            value = self._prompt_cli_str(shared_ctx)
        return value

    def _prompt_cli_str(self, shared_ctx: AnySharedContext) -> str:
        prompt_message = self.prompt_message
        default_value = self._get_default_str(shared_ctx)
        if default_value != "":
            prompt_message = f"{prompt_message} [{default_value}]"
        value = input(f"{prompt_message}: ")
        if value.strip() == "":
            value = default_value
        return value

    def _get_default_str(self, shared_ctx: AnySharedContext) -> str:
        return get_str_attr(
            shared_ctx, self._default_str, auto_render=self._auto_render
        )



import os

from zrb.group.any_group import AnyGroup
from zrb.session.any_session import AnySession
from zrb.task.any_task import AnyTask
from zrb.util.string.format import fstring_format

_DIR = os.path.dirname(__file__)

with open(os.path.join(_DIR, "view.html"), "r") as f:
    _VIEW_TEMPLATE = f.read()

with open(os.path.join(_DIR, "partial", "input.html")) as f:
    _TASK_INPUT_TEMPLATE = f.read()

with open(os.path.join(_DIR, "partial", "main.js")) as f:
    _MAIN_SCRIPT = f.read()

with open(os.path.join(_DIR, "partial", "show-existing-session.js")) as f:
    _SHOW_EXISTING_SESSION_SCRIPT = f.read()

with open(os.path.join(_DIR, "partial", "visualize-history.js")) as f:
    _VISUALIZE_HISTORY_SCRIPT = f.read()

with open(os.path.join(_DIR, "partial", "common-util.js")) as f:
    _COMMON_UTIL_SCRIPT = f.read()


def handle_task_ui(
    root_group: AnyGroup, task: AnyTask, session: AnySession, url: str, args: list[str]
):
    from fastapi.responses import HTMLResponse

    session.register_task(task)
    ctx = task.get_ctx(session)
    url_parts = url.split("/")
    # Assemble parent url
    parent_url_parts = url_parts[:-2] + [""]
    parent_url = "/".join(parent_url_parts)
    # Assemble api url
    api_url_parts = list(url_parts)
    api_url_parts[1] = "api"
    api_url = "/".join(api_url_parts)
    # Assemble ui url
    ui_url_parts = list(api_url_parts)
    ui_url_parts[1] = "ui"
    ui_url = "/".join(ui_url_parts)
    task_inputs = "\n".join(
        [
            fstring_format(_TASK_INPUT_TEMPLATE, {"task_input": task_input, "ctx": ctx})
            for task_input in task.inputs
        ]
    )
    session_name = args[0] if len(args) > 0 else ""
    return HTMLResponse(
        fstring_format(
            _VIEW_TEMPLATE,
            {
                "name": task.name,
                "description": task.description,
                "root_name": root_group.name,
                "root_description": root_group.description,
                "url": url,
                "parent_url": parent_url,
                "task_inputs": task_inputs,
                "api_url": api_url,
                "ui_url": ui_url,
                "main_script": _MAIN_SCRIPT,
                "show_existing_session_script": _SHOW_EXISTING_SESSION_SCRIPT,
                "visualize_history_script": _VISUALIZE_HISTORY_SCRIPT,
                "common_util_script": _COMMON_UTIL_SCRIPT,
                "session_name": session_name,
            },
        )
    )



import datetime
from abc import ABC, abstractmethod

from zrb.session_state_log.session_state_log import SessionStateLog, SessionStateLogList


class AnySessionStateLogger(ABC):

    @abstractmethod
    def write(self, session_log: SessionStateLog):
        pass

    @abstractmethod
    def read(self, session_name: str) -> SessionStateLog:
        pass

    @abstractmethod
    def list(
        self,
        task_path: list[str],
        min_start_time: datetime.datetime,
        max_start_time: datetime.datetime,
        page: int = 0,
        limit: int = 10,
    ) -> SessionStateLogList:
        pass



from typing import Any, Callable, Generic, Type, TypeVar

from fastapp_template.common.error import NotFoundError
from sqlalchemy import Engine
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
from sqlmodel import Session, SQLModel, select

DBModel = TypeVar("DBModel", bound=SQLModel)
ResponseModel = TypeVar("Model", bound=SQLModel)
CreateModel = TypeVar("CreateModel", bound=SQLModel)
UpdateModel = TypeVar("UpdateModel", bound=SQLModel)


class BaseDBRepository(Generic[DBModel, ResponseModel, CreateModel, UpdateModel]):
    db_model: Type[DBModel]
    response_model: Type[ResponseModel]
    create_model: Type[CreateModel]
    update_model: Type[UpdateModel]
    entity_name: str = "entity"
    column_preprocessors: dict[str, Callable[[Any], Any]] = {}

    def __init__(self, engine: Engine | AsyncEngine):
        self.engine = engine
        self.is_async = isinstance(engine, AsyncEngine)

    def _to_response(self, db_instance: DBModel) -> ResponseModel:
        return self.response_model(**db_instance.model_dump())

    async def create(self, data: CreateModel) -> ResponseModel:
        data_dict = data.model_dump(exclude_unset=True)
        for key, preprocessor in self.column_preprocessors.items():
            if key in data_dict:
                data_dict[key] = preprocessor(data_dict[key])
        db_instance = self.db_model(**data_dict)
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                session.add(db_instance)
                await session.commit()
                await session.refresh(db_instance)
        else:
            with Session(self.engine) as session:
                session.add(db_instance)
                session.commit()
                session.refresh(db_instance)
        return self._to_response(db_instance)

    async def get_by_id(self, item_id: str) -> ResponseModel:
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                db_instance = await session.get(self.db_model, item_id)
        else:
            with Session(self.engine) as session:
                db_instance = session.get(self.db_model, item_id)
        if not db_instance:
            raise NotFoundError(f"{self.entity_name} not found")
        return self._to_response(db_instance)

    async def get_all(self, page: int = 1, page_size: int = 10) -> list[ResponseModel]:
        offset = (page - 1) * page_size
        statement = select(self.db_model).offset(offset).limit(page_size)
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                result = await session.execute(statement)
                results = result.scalars().all()
        else:
            with Session(self.engine) as session:
                results = session.exec(statement).all()
        return [self._to_response(instance) for instance in results]

    async def update(self, item_id: str, data: UpdateModel) -> ResponseModel:
        update_data = data.model_dump(exclude_unset=True)
        for key, value in update_data.items():
            if key in self.column_preprocessors:
                update_data[key] = self.column_preprocessors[key](value)
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                db_instance = await session.get(self.db_model, item_id)
                if not db_instance:
                    raise NotFoundError(f"{self.entity_name} not found")
                for key, value in update_data.items():
                    setattr(db_instance, key, value)
                session.add(db_instance)
                await session.commit()
                await session.refresh(db_instance)
        else:
            with Session(self.engine) as session:
                db_instance = session.get(self.db_model, item_id)
                if not db_instance:
                    raise NotFoundError(f"{self.entity_name} not found")
                for key, value in update_data.items():
                    setattr(db_instance, key, value)
                session.add(db_instance)
                session.commit()
                session.refresh(db_instance)
        return self._to_response(db_instance)

    async def delete(self, item_id: str) -> ResponseModel:
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                db_instance = await session.get(self.db_model, item_id)
                if not db_instance:
                    raise NotFoundError(f"{self.entity_name} not found")
                await session.delete(db_instance)
                await session.commit()
        else:
            with Session(self.engine) as session:
                db_instance = session.get(self.db_model, item_id)
                if not db_instance:
                    raise NotFoundError(f"{self.entity_name} not found")
                session.delete(db_instance)
                session.commit()
        return self._to_response(db_instance)

    async def create_bulk(self, data_list: list[CreateModel]) -> list[ResponseModel]:
        db_instances = []
        for data in data_list:
            data_dict = data.model_dump(exclude_unset=True)
            for key, preprocessor in self.column_preprocessors.items():
                if key in data_dict:
                    data_dict[key] = preprocessor(data_dict[key])
            db_instances.append(self.db_model(**data_dict))
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                session.add_all(db_instances)
                await session.commit()
                for instance in db_instances:
                    await session.refresh(instance)
        else:
            with Session(self.engine) as session:
                session.add_all(db_instances)
                session.commit()
                for instance in db_instances:
                    session.refresh(instance)
        return [self._to_response(instance) for instance in db_instances]



import asyncio
import os
import sys
from datetime import datetime, timedelta
from typing import Any, Dict, List

from zrb.config import BANNER, WEB_HTTP_PORT
from zrb.context.shared_context import SharedContext
from zrb.group.any_group import AnyGroup
from zrb.runner.web_controller.group_info_ui.controller import handle_group_info_ui
from zrb.runner.web_controller.home_page.controller import handle_home_page
from zrb.runner.web_controller.task_ui.controller import handle_task_ui
from zrb.runner.web_util import NewSessionResponse
from zrb.session.session import Session
from zrb.session_state_log.session_state_log import SessionStateLog, SessionStateLogList
from zrb.session_state_logger.default_session_state_logger import (
    default_session_state_logger,
)
from zrb.task.any_task import AnyTask
from zrb.util.group import extract_node_from_args, get_node_path


def create_app(root_group: AnyGroup, port: int = WEB_HTTP_PORT):
    from contextlib import asynccontextmanager

    from fastapi import FastAPI, HTTPException, Request
    from fastapi.responses import FileResponse, HTMLResponse
    from fastapi.staticfiles import StaticFiles

    _STATIC_DIR = os.path.join(os.path.dirname(__file__), "web_controller", "static")
    _COROS = []

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        for line in BANNER.split("\n") + [
            f"Zrb Server running on http://localhost:{port}"
        ]:
            print(line, file=sys.stderr)
        yield
        for coro in _COROS:
            coro.cancel()
        asyncio.gather(*_COROS)

    app = FastAPI(title="zrb", lifespan=lifespan)

    # Serve static files
    app.mount("/static", StaticFiles(directory=_STATIC_DIR), name="static")

    @app.get("/", response_class=HTMLResponse, include_in_schema=False)
    @app.get("/ui", response_class=HTMLResponse, include_in_schema=False)
    @app.get("/ui/", response_class=HTMLResponse, include_in_schema=False)
    async def home_page():
        return handle_home_page(root_group)

    @app.get("/static/{file_path:path}", include_in_schema=False)
    async def static_files(file_path: str):
        full_path = os.path.join(_STATIC_DIR, file_path)
        if os.path.isfile(full_path):
            return FileResponse(full_path)
        raise HTTPException(status_code=404, detail="File not found")

    @app.get("/ui/{path:path}", include_in_schema=False)
    async def ui_page(path: str):
        # Avoid capturing '/ui' itself
        if not path:
            raise HTTPException(status_code=404, detail="Not Found")
        args = path.split("/")
        node, node_path, residual_args = extract_node_from_args(root_group, args)
        url = f"/ui/{'/'.join(node_path)}/"
        if isinstance(node, AnyTask):
            shared_ctx = SharedContext(env=dict(os.environ))
            session = Session(shared_ctx=shared_ctx, root_group=root_group)
            return handle_task_ui(root_group, node, session, url, residual_args)
        elif isinstance(node, AnyGroup):
            return handle_group_info_ui(root_group, node, url)
        raise HTTPException(status_code=404, detail="Not Found")

    @app.post("/api/{path:path}")
    async def create_new_session(
        path: str, request: Request = None
    ) -> NewSessionResponse:
        """
        Creating new session
        """
        args = path.split("/")
        node, _, residual_args = extract_node_from_args(root_group, args)
        if isinstance(node, AnyTask):
            session_name = residual_args[0] if residual_args else None
            if not session_name:
                body = await request.json()
                shared_ctx = SharedContext(env=dict(os.environ))
                session = Session(shared_ctx=shared_ctx, root_group=root_group)
                coro = asyncio.create_task(node.async_run(session, str_kwargs=body))
                _COROS.append(coro)
                coro.add_done_callback(lambda coro: _COROS.remove(coro))
                return NewSessionResponse(session_name=session.name)
        raise HTTPException(status_code=404, detail="Not Found")

    @app.get("/api/{path:path}", response_model=SessionStateLog | SessionStateLogList)
    async def get_session(path: str, query_params: Dict[str, Any] = {}):
        """
        Getting existing session or sessions
        """
        args = path.split("/")
        node, _, residual_args = extract_node_from_args(root_group, args)
        if isinstance(node, AnyTask) and residual_args:
            if residual_args[0] == "list":
                task_path = get_node_path(root_group, node)
                return list_sessions(task_path, query_params)
            else:
                return read_session(residual_args[0])
        raise HTTPException(status_code=404, detail="Not Found")

    def list_sessions(
        task_path: List[str], query_params: Dict[str, Any]
    ) -> SessionStateLogList:
        max_start_time = datetime.now()
        if "to" in query_params:
            max_start_time = datetime.strptime(query_params["to"], "%Y-%m-%d %H:%M:%S")
        min_start_time = max_start_time - timedelta(hours=1)
        if "from" in query_params:
            min_start_time = datetime.strptime(
                query_params["from"], "%Y-%m-%d %H:%M:%S"
            )
        page = int(query_params.get("page", 0))
        limit = int(query_params.get("limit", 10))
        try:
            return default_session_state_logger.list(
                task_path,
                min_start_time=min_start_time,
                max_start_time=max_start_time,
                page=page,
                limit=limit,
            )
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    def read_session(session_name: str) -> SessionStateLog:
        try:
            return default_session_state_logger.read(session_name)
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))

    return app



"""Add user table

Revision ID: 3093c7336477
Revises:
Create Date: 2024-11-20 05:57:01.684118

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "3093c7336477"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user",
        sa.Column("id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("username", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("password", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("user")
    # ### end Alembic commands ###



from zrb.group.any_group import AnyGroup
from zrb.task.any_task import AnyTask


class NodeNotFoundError(ValueError):
    pass


def extract_node_from_args(
    root_group: AnyGroup, args: list[str], web_only: bool = False
) -> tuple[AnyGroup | AnyTask, list[str], list[str]]:
    node = root_group
    node_path = []
    residual_args = []
    for index, name in enumerate(args):
        task = node.get_task_by_alias(name)
        if web_only and task is not None and task.cli_only:
            task = None
        group = node.get_group_by_alias(name)
        if group is not None and len(get_all_subtasks(group, web_only)) == 0:
            # If group doesn't contain any task, then ignore its existence
            group = None
        if task is None and group is None:
            raise NodeNotFoundError(
                f"Invalid subcommand: {root_group.name} {' '.join(args)}"
            )
        node_path.append(name)
        if group is not None:
            if task is not None and index == len(args) - 1:
                node = task
                residual_args = args[index + 1 :]
                break
            node = group
            continue
        if task is not None:
            node = task
            residual_args = args[index + 1 :]
            break
    return node, node_path, residual_args


def get_node_path(group: AnyGroup, node: AnyGroup | AnyTask) -> list[str] | None:
    if group is None:
        return []
    if isinstance(node, AnyTask):
        for alias, subtask in group.subtasks.items():
            if subtask == node:
                return [alias]
    if isinstance(node, AnyGroup):
        for alias, subgroup in group.subgroups.items():
            if subgroup == node:
                return [alias]
    for alias, subgroup in group.subgroups.items():
        result = get_node_path(subgroup, node)
        if result is not None:
            return [alias] + result
    return None


def get_non_empty_subgroups(
    group: AnyGroup, web_only: bool = False
) -> dict[str, AnyGroup]:
    return {
        alias: subgroup
        for alias, subgroup in group.subgroups.items()
        if len(get_all_subtasks(subgroup, web_only)) > 0
    }


def get_subtasks(group: AnyGroup, web_only: bool = False) -> dict[str, AnyTask]:
    return {
        alias: subtask
        for alias, subtask in group.subtasks.items()
        if not web_only or (web_only and not subtask.cli_only)
    }


def get_all_subtasks(group: AnyGroup, web_only: bool = False) -> list[AnyTask]:
    subtasks = [
        subtask
        for subtask in group.subtasks.values()
        if not web_only or (web_only and not subtask.cli_only)
    ]
    for subgroup in group.subgroups.values():
        subtasks += get_all_subtasks(subgroup, web_only)
    return subtasks



from contextlib import asynccontextmanager

from fastapi import FastAPI
from fastapp_template.common.db_engine import engine
from fastapp_template.config import APP_MODE, APP_MODULES
from sqlmodel import SQLModel


@asynccontextmanager
async def lifespan(app: FastAPI):
    SQLModel.metadata.create_all(engine)
    yield


app_title = (
    "App Name" if APP_MODE == "monolith" else f"App Name - {', '.join(APP_MODULES)}"
)
app = FastAPI(title=app_title, lifespan=lifespan)



import asyncio
import os
from collections.abc import Callable
from typing import Any

from zrb.attr.type import BoolAttr, fstring
from zrb.context.any_context import AnyContext
from zrb.context.shared_context import AnySharedContext, SharedContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.session.any_session import AnySession
from zrb.session.session import Session
from zrb.task.any_task import AnyTask
from zrb.util.attr import get_bool_attr
from zrb.util.run import run_async
from zrb.xcom.xcom import Xcom


class BaseTask(AnyTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput | None] | AnyInput | None = None,
        env: list[AnyEnv | None] | AnyEnv | None = None,
        action: fstring | Callable[[AnyContext], Any] | None = None,
        execute_condition: BoolAttr = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
        successor: list[AnyTask] | AnyTask | None = None,
    ):
        self._name = name
        self._color = color
        self._icon = icon
        self._description = description
        self._cli_only = cli_only
        self._inputs = input
        self._envs = env
        self._retries = retries
        self._retry_period = retry_period
        self._upstreams = upstream
        self._fallbacks = fallback
        self._successors = successor
        self._readiness_checks = readiness_check
        self._readiness_check_delay = readiness_check_delay
        self._readiness_check_period = readiness_check_period
        self._readiness_failure_threshold = readiness_failure_threshold
        self._readiness_timeout = readiness_timeout
        self._monitor_readiness = monitor_readiness
        self._execute_condition = execute_condition
        self._action = action

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self._name}>"

    def __rshift__(self, other: AnyTask | list[AnyTask]) -> AnyTask:
        try:
            if isinstance(other, AnyTask):
                other.append_upstream(self)
            elif isinstance(other, list):
                for task in other:
                    task.append_upstream(self)
            return other
        except Exception as e:
            raise ValueError(f"Invalid operation {self} >> {other}: {e}")

    def __lshift__(self, other: AnyTask | list[AnyTask]) -> AnyTask:
        try:
            self.append_upstream(other)
            return self
        except Exception as e:
            raise ValueError(f"Invalid operation {self} << {other}: {e}")

    @property
    def name(self) -> str:
        return self._name

    @property
    def color(self) -> int | None:
        return self._color

    @property
    def icon(self) -> str | None:
        return self._icon

    @property
    def description(self) -> str:
        return self._description if self._description is not None else self.name

    @property
    def cli_only(self) -> bool:
        return self._cli_only

    @property
    def envs(self) -> list[AnyEnv]:
        envs = []
        for upstream in self.upstreams:
            envs += upstream.envs
        if isinstance(self._envs, AnyEnv):
            envs.append(self._envs)
        elif self._envs is not None:
            envs += self._envs
        return [env for env in envs if env is not None]

    @property
    def inputs(self) -> list[AnyInput]:
        inputs = []
        for upstream in self.upstreams:
            self.__combine_inputs(inputs, upstream.inputs)
        if self._inputs is not None:
            self.__combine_inputs(inputs, self._inputs)
        return [task_input for task_input in inputs if task_input is not None]

    def __combine_inputs(
        self, inputs: list[AnyInput], other_inputs: list[AnyInput] | AnyInput
    ):
        input_names = [task_input.name for task_input in inputs]
        if isinstance(other_inputs, AnyInput):
            other_inputs = [other_inputs]
        elif other_inputs is None:
            other_inputs = []
        for task_input in other_inputs:
            if task_input is None:
                continue
            if task_input.name not in input_names:
                inputs.append(task_input)

    @property
    def fallbacks(self) -> list[AnyTask]:
        if self._fallbacks is None:
            return []
        elif isinstance(self._fallbacks, AnyTask):
            return [self._fallbacks]
        return self._fallbacks

    def append_fallback(self, fallbacks: AnyTask | list[AnyTask]):
        fallback_list = [fallbacks] if isinstance(fallbacks, AnyTask) else fallbacks
        for fallback in fallback_list:
            self.__append_fallback(fallback)

    def __append_fallback(self, fallback: AnyTask):
        # Make sure self._fallbacks is a list
        if self._fallbacks is None:
            self._fallbacks = []
        elif isinstance(self._fallbacks, AnyTask):
            self._fallbacks = [self._fallbacks]
        # Add fallback if it was not on self._fallbacks
        if fallback not in self._fallbacks:
            self._fallbacks.append(fallback)

    @property
    def successors(self) -> list[AnyTask]:
        if self._successors is None:
            return []
        elif isinstance(self._successors, AnyTask):
            return [self._successors]
        return self._successors

    def append_successor(self, successors: AnyTask | list[AnyTask]):
        successor_list = [successors] if isinstance(successors, AnyTask) else successors
        for successor in successor_list:
            self.__append_successor(successor)

    def __append_successor(self, successor: AnyTask):
        # Make sure self._successors is a list
        if self._successors is None:
            self._successors = []
        elif isinstance(self._successors, AnyTask):
            self._successors = [self._successors]
        # Add successor if it was not on self._successors
        if successor not in self._successors:
            self._successors.append(successor)

    @property
    def readiness_checks(self) -> list[AnyTask]:
        if self._readiness_checks is None:
            return []
        elif isinstance(self._readiness_checks, AnyTask):
            return [self._readiness_checks]
        return self._readiness_checks

    def append_readiness_check(self, readiness_checks: AnyTask | list[AnyTask]):
        readiness_check_list = (
            [readiness_checks]
            if isinstance(readiness_checks, AnyTask)
            else readiness_checks
        )
        for readiness_check in readiness_check_list:
            self.__append_readiness_check(readiness_check)

    def __append_readiness_check(self, readiness_check: AnyTask):
        # Make sure self._readiness_checks is a list
        if self._readiness_checks is None:
            self._readiness_checks = []
        elif isinstance(self._readiness_checks, AnyTask):
            self._readiness_checks = [self._readiness_checks]
        # Add readiness_check if it was not on self._readiness_checks
        if readiness_check not in self._readiness_checks:
            self._readiness_checks.append(readiness_check)

    @property
    def upstreams(self) -> list[AnyTask]:
        if self._upstreams is None:
            return []
        elif isinstance(self._upstreams, AnyTask):
            return [self._upstreams]
        return self._upstreams

    def append_upstream(self, upstreams: AnyTask | list[AnyTask]):
        upstream_list = [upstreams] if isinstance(upstreams, AnyTask) else upstreams
        for upstream in upstream_list:
            self.__append_upstream(upstream)

    def __append_upstream(self, upstream: AnyTask):
        # Make sure self._upstreams is a list
        if self._upstreams is None:
            self._upstreams = []
        elif isinstance(self._upstreams, AnyTask):
            self._upstreams = [self._upstreams]
        # Add upstream if it was not on self._upstreams
        if upstream not in self._upstreams:
            self._upstreams.append(upstream)

    def get_ctx(self, session: AnySession) -> AnyContext:
        ctx = session.get_ctx(self)
        # Enhance session ctx with current task env
        for env in self.envs:
            env.update_context(ctx)
        return ctx

    def run(
        self, session: AnySession | None = None, str_kwargs: dict[str, str] = {}
    ) -> Any:
        return asyncio.run(self.async_run(session, str_kwargs))

    async def async_run(
        self, session: AnySession | None = None, str_kwargs: dict[str, str] = {}
    ) -> Any:
        if session is None:
            session = Session(shared_ctx=SharedContext())
        # Update session
        self.__fill_shared_context_inputs(session.shared_ctx, str_kwargs)
        self.__fill_shared_context_envs(session.shared_ctx)
        return await run_async(self.exec_root_tasks(session))

    def __fill_shared_context_inputs(
        self, shared_context: AnySharedContext, str_kwargs: dict[str, str] = {}
    ):
        for task_input in self.inputs:
            if task_input.name not in shared_context.input:
                str_value = str_kwargs.get(task_input.name, None)
                task_input.update_shared_context(shared_context, str_value)

    def __fill_shared_context_envs(self, shared_context: AnySharedContext):
        # Inject os environ
        os_env_map = {
            key: val
            for key, val in os.environ.items()
            if key not in shared_context._env
        }
        shared_context._env.update(os_env_map)

    async def exec_root_tasks(self, session: AnySession):
        session.set_main_task(self)
        session.state_logger.write(session.as_state_log())
        log_state = asyncio.create_task(self._log_session_state(session))
        root_tasks = [
            task
            for task in session.get_root_tasks(self)
            if session.is_allowed_to_run(task)
        ]
        root_task_coros = [
            run_async(root_task.exec_chain(session)) for root_task in root_tasks
        ]
        try:
            await asyncio.gather(*root_task_coros)
            await session.wait_deferred()
            session.terminate()
            await log_state
            return session.final_result
        except IndexError:
            return None
        except asyncio.CancelledError:
            ctx = self.get_ctx(session)
            ctx.log_info("Session terminated")
        finally:
            session.terminate()
            session.state_logger.write(session.as_state_log())
            ctx = self.get_ctx(session)
            ctx.log_debug(session)

    async def _log_session_state(self, session: AnySession):
        session.state_logger.write(session.as_state_log())
        while not session.is_terminated:
            await asyncio.sleep(0.1)
            session.state_logger.write(session.as_state_log())
        session.state_logger.write(session.as_state_log())

    async def exec_chain(self, session: AnySession):
        if session.is_terminated or not session.is_allowed_to_run(self):
            return
        result = await self.exec(session)
        # Get next tasks
        nexts = session.get_next_tasks(self)
        if session.is_terminated or len(nexts) == 0:
            return result
        # Run next tasks asynchronously
        next_coros = [run_async(next.exec_chain(session)) for next in nexts]
        return await asyncio.gather(*next_coros)

    async def exec(self, session: AnySession):
        ctx = self.get_ctx(session)
        if not session.is_allowed_to_run(self):
            # Task is not allowed to run, skip it for now.
            # This will be triggered later
            ctx.log_info("Not allowed to run")
            return
        if not self.__get_execute_condition(session):
            # Skip the task
            ctx.log_info("Marked as skipped")
            session.get_task_status(self).mark_as_skipped()
            return
        # Wait for task to be ready
        await run_async(self.__exec_action_until_ready(session))

    def __get_execute_condition(self, session: Session) -> bool:
        ctx = self.get_ctx(session)
        return get_bool_attr(ctx, self._execute_condition, True, auto_render=True)

    async def __exec_action_until_ready(self, session: AnySession):
        ctx = self.get_ctx(session)
        readiness_checks = self.readiness_checks
        if len(readiness_checks) == 0:
            ctx.log_info("No readiness checks")
            # Task has no readiness check
            result = await run_async(self.__exec_action_and_retry(session))
            ctx.log_info("Marked as ready")
            session.get_task_status(self).mark_as_ready()
            return result
        # Start the task along with the readiness checks
        action_coro = asyncio.create_task(
            run_async(self.__exec_action_and_retry(session))
        )
        await asyncio.sleep(self._readiness_check_delay)
        readiness_check_coros = [
            run_async(check.exec_chain(session)) for check in readiness_checks
        ]
        # Only wait for readiness checks and mark the task as ready
        ctx.log_info("Start readiness checks")
        result = await asyncio.gather(*readiness_check_coros)
        ctx.log_info("Readiness checks completed")
        ctx.log_info("Marked as ready")
        session.get_task_status(self).mark_as_ready()
        # Defer task's coroutines, will be waited later
        session.defer_action(self, action_coro)
        if self._monitor_readiness:
            monitor_and_rerun_coro = asyncio.create_task(
                run_async(self.__exec_monitoring(session, action_coro))
            )
            session.defer_monitoring(self, monitor_and_rerun_coro)
        return result

    async def __exec_monitoring(self, session: AnySession, action_coro: asyncio.Task):
        readiness_checks = self.readiness_checks
        failure_count = 0
        ctx = self.get_ctx(session)
        while not session.is_terminated:
            await asyncio.sleep(self._readiness_check_period)
            if failure_count < self._readiness_failure_threshold:
                for readiness_check in readiness_checks:
                    session.get_task_status(readiness_check).reset_history()
                    session.get_task_status(readiness_check).reset()
                    readiness_xcom: Xcom = ctx.xcom[self.name]
                    readiness_xcom.clear()
                readiness_check_coros = [
                    check.exec_chain(session) for check in readiness_checks
                ]
                try:
                    ctx.log_info("Checking")
                    await asyncio.wait_for(
                        asyncio.gather(*readiness_check_coros),
                        timeout=self._readiness_timeout,
                    )
                    ctx.log_info("OK")
                    continue
                except (asyncio.CancelledError, KeyboardInterrupt):
                    for readiness_check in readiness_checks:
                        ctx.log_info("Marked as failed")
                        session.get_task_status(readiness_check).mark_as_failed()
                except asyncio.TimeoutError:
                    failure_count += 1
                    ctx.log_info("Detecting failure")
                    ctx.log_debug(f"Failure count: {failure_count}")
            # Readiness check failed, reset
            ctx.log_info("Resetting")
            action_coro.cancel()
            session.get_task_status(self).reset()
            # defer this action
            ctx.log_info("Running")
            action_coro = asyncio.create_task(
                run_async(self.__exec_action_and_retry(session))
            )
            session.defer_action(self, action_coro)
            failure_count = 0
            ctx.log_info("Continue monitoring")

    async def __exec_action_and_retry(self, session: AnySession) -> Any:
        ctx = self.get_ctx(session)
        max_attempt = self._retries + 1
        ctx.set_max_attempt(max_attempt)
        for attempt in range(max_attempt):
            ctx.set_attempt(attempt + 1)
            if attempt > 0:
                # apply retry period only if this is not the first attempt
                await asyncio.sleep(self._retry_period)
            try:
                ctx.log_info("Marked as started")
                session.get_task_status(self).mark_as_started()
                result = await run_async(self._exec_action(ctx))
                ctx.log_info("Marked as completed")
                session.get_task_status(self).mark_as_completed()
                # Put result on xcom
                task_xcom: Xcom = ctx.xcom.get(self.name)
                task_xcom.push(result)
                await run_async(self.__exec_successors(session))
                return result
            except (asyncio.CancelledError, KeyboardInterrupt):
                ctx.log_info("Marked as failed")
                session.get_task_status(self).mark_as_failed()
                return
            except Exception as e:
                ctx.log_error(e)
                if attempt < max_attempt - 1:
                    ctx.log_info("Marked as failed")
                    session.get_task_status(self).mark_as_failed()
                    continue
                ctx.log_info("Marked as permanently failed")
                session.get_task_status(self).mark_as_permanently_failed()
                await run_async(self.__exec_fallbacks(session))
                raise e

    async def __exec_successors(self, session: AnySession) -> Any:
        successors: list[AnyTask] = self.successors
        successor_coros = [
            run_async(successor.exec_chain(session)) for successor in successors
        ]
        await asyncio.gather(*successor_coros)

    async def __exec_fallbacks(self, session: AnySession) -> Any:
        fallbacks: list[AnyTask] = self.fallbacks
        fallback_coros = [
            run_async(fallback.exec_chain(session)) for fallback in fallbacks
        ]
        await asyncio.gather(*fallback_coros)

    async def _exec_action(self, ctx: AnyContext) -> Any:
        """Execute the main action of the task.
        By default will render and run the _action attribute.

        Override this method to define custom action.

        Args:
            session (AnySession): The shared session.

        Returns:
            Any: The result of the action execution.
        """
        if self._action is None:
            return
        if isinstance(self._action, str):
            return ctx.render(self._action)
        return await run_async(self._action(ctx))



from zrb.builtin.group import llm_group
from zrb.builtin.llm.tool.cli import run_shell_command
from zrb.builtin.llm.tool.web import open_web_page, query_internet
from zrb.config import (
    LLM_ALLOW_ACCESS_SHELL,
    LLM_ALLOW_ACCESS_WEB,
    LLM_HISTORY_FILE,
    LLM_MODEL,
    LLM_SYSTEM_PROMPT,
)
from zrb.input.str_input import StrInput
from zrb.input.text_input import TextInput
from zrb.task.llm_task import LLMTask

llm_chat: LLMTask = llm_group.add_task(
    LLMTask(
        name="llm-chat",
        input=[
            StrInput(
                "model",
                description="LLM Model",
                prompt="LLM Model",
                default_str=LLM_MODEL,
            ),
            StrInput(
                "system-prompt",
                description="System prompt",
                prompt="System prompt",
                default_str=LLM_SYSTEM_PROMPT,
            ),
            TextInput("message", description="User message", prompt="Your message"),
        ],
        history_file=LLM_HISTORY_FILE,
        description="Chat with LLM",
        model="{ctx.input.model}",
        system_prompt="{ctx.input['system-prompt']}",
        message="{ctx.input.message}",
    ),
    alias="chat",
)

if LLM_ALLOW_ACCESS_SHELL:
    llm_chat.add_tool(run_shell_command)

if LLM_ALLOW_ACCESS_WEB:
    llm_chat.add_tool(open_web_page)
    llm_chat.add_tool(query_internet)



import importlib
import importlib.util
import os
import re
import sys
from functools import lru_cache
from typing import Any

pattern = re.compile("[^a-zA-Z0-9]")


def load_zrb_init(dir_path: str | None = None):
    if dir_path is None:
        dir_path = os.getcwd()
    script_path = os.path.join(dir_path, "zrb_init.py")
    if os.path.isfile(script_path):
        load_file(script_path, -1)
        return
    new_dir_path = os.path.dirname(dir_path)
    if new_dir_path == dir_path:
        return
    load_zrb_init(new_dir_path)


@lru_cache()
def load_file(script_path: str, sys_path_index: int = 0):
    if not os.path.isfile(script_path):
        return
    # Append script dir path
    script_dir_path = os.path.dirname(script_path)
    if script_dir_path not in sys.path:
        sys.path.insert(sys_path_index, script_dir_path)
    # Add script dir path to Python path
    os.environ["PYTHONPATH"] = _get_new_python_path(script_dir_path)
    module_name = pattern.sub("", script_path)
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)


def _get_new_python_path(dir_path: str) -> str:
    current_python_path = os.getenv("PYTHONPATH")
    if current_python_path is None or current_python_path == "":
        return dir_path
    if dir_path in current_python_path.split(":"):
        return current_python_path
    return ":".join([current_python_path, dir_path])


@lru_cache()
def load_module(module_name: str) -> Any:
    module = importlib.import_module(module_name)
    return module



from fastapp_template.common.app import app
from fastapp_template.common.schema import BasicResponse
from fastapp_template.config import APP_MODE, APP_MODULES
from fastapp_template.module.auth.service.user.usecase import user_usecase

if APP_MODE == "microservices" and "auth" in APP_MODULES:

    if APP_MODE == "microservices" and (
        len(APP_MODULES) > 0 and APP_MODULES[0] == "auth"
    ):

        @app.api_route("/health", methods=["GET", "HEAD"], response_model=BasicResponse)
        async def health():
            return BasicResponse(message="ok")

        @app.api_route(
            "/readiness", methods=["GET", "HEAD"], response_model=BasicResponse
        )
        async def readiness():
            return BasicResponse(message="ok")

    user_usecase.serve_route(app)



import re


def check_unrecommended_commands(cmd_script: str) -> dict[str, str]:
    banned_commands = {
        "<(": "Process substitution isn't POSIX compliant and causes trouble",
        "column": "Command isn't included in Ubuntu packages and is not POSIX compliant",
        "echo": "echo isn't consistent across OS; use printf instead",
        "eval": "Avoid eval as it can accidentally execute arbitrary strings",
        "realpath": "Not available by default on OSX",
        "source": "Not POSIX compliant; use '.' instead",
        " test": "Use '[' instead for consistency",
        "which": "Command in not POSIX compliant, use command -v",
    }
    banned_commands_regex = {
        r"grep.* -y": "grep -y does not work on Alpine; use grep -i",
        r"grep.* -P": "grep -P is not valid on OSX",
        r"grep[^|]+--\w{2,}": "grep long commands do not work on Alpine",
        r'readlink.+-.*f.+["$]': "readlink -f behaves differently on OSX",
        r"sort.*-V": "sort -V is not supported everywhere",
        r"sort.*--sort-versions": "sort --sort-version is not supported everywhere",
        r"\bls ": "Avoid using ls; use shell globs or find instead",
    }
    violations = {}
    # Check banned commands
    for cmd, reason in banned_commands.items():
        if cmd in cmd_script:
            violations[cmd] = reason
    # Check banned regex patterns
    for pattern, reason in banned_commands_regex.items():
        if re.search(pattern, cmd_script):
            violations[pattern] = reason
    return violations



from pydantic import BaseModel


class NewSessionResponse(BaseModel):
    session_name: str


def url_to_args(url: str) -> list[str]:
    stripped_url = url.strip("/")
    return [part for part in stripped_url.split("/") if part.strip() != ""]


def node_path_to_url(args: list[str]) -> str:
    pruned_args = [part for part in args if part.strip() != ""]
    stripped_url = "/".join(pruned_args)
    return f"/{stripped_url}/"



import json
import os
import sys
from collections.abc import Callable, Iterable

import litellm

from zrb.config import (
    RAG_CHUNK_SIZE,
    RAG_EMBEDDING_MODEL,
    RAG_MAX_RESULT_COUNT,
    RAG_OVERLAP,
)
from zrb.util.cli.style import stylize_error, stylize_faint
from zrb.util.run import run_async

Document = str | Callable[[], str]
Documents = Callable[[], Iterable[Document]] | Iterable[Document]


def create_rag_from_directory(
    tool_name: str,
    tool_description: str,
    document_dir_path: str = "./documents",
    model: str = RAG_EMBEDDING_MODEL,
    vector_db_path: str = "./chroma",
    vector_db_collection: str = "documents",
    chunk_size: int = RAG_CHUNK_SIZE,
    overlap: int = RAG_OVERLAP,
    max_result_count: int = RAG_MAX_RESULT_COUNT,
):
    return create_rag(
        tool_name=tool_name,
        tool_description=tool_description,
        documents=get_rag_documents(os.path.expanduser(document_dir_path)),
        model=model,
        vector_db_path=vector_db_path,
        vector_db_collection=vector_db_collection,
        reset_db=get_rag_reset_db(
            document_dir_path=os.path.expanduser(document_dir_path),
            vector_db_path=os.path.expanduser(vector_db_path),
        ),
        chunk_size=chunk_size,
        overlap=overlap,
        max_result_count=max_result_count,
    )


def create_rag(
    tool_name: str,
    tool_description: str,
    documents: Documents = [],
    model: str = RAG_EMBEDDING_MODEL,
    vector_db_path: str = "./chroma",
    vector_db_collection: str = "documents",
    reset_db: Callable[[], bool] | bool = False,
    chunk_size: int = RAG_CHUNK_SIZE,
    overlap: int = RAG_OVERLAP,
    max_result_count: int = RAG_MAX_RESULT_COUNT,
) -> Callable[[str], str]:
    async def retrieve(query: str) -> str:
        import chromadb
        from chromadb.config import Settings

        is_db_exist = os.path.isdir(vector_db_path)
        client = chromadb.PersistentClient(
            path=vector_db_path, settings=Settings(allow_reset=True)
        )
        should_reset_db = (
            await run_async(reset_db()) if callable(reset_db) else reset_db
        )
        if (not is_db_exist) or should_reset_db:
            client.reset()
            collection = client.get_or_create_collection(vector_db_collection)
            chunk_index = 0
            print(stylize_faint("Scanning documents"), file=sys.stderr)
            docs = await run_async(documents()) if callable(documents) else documents
            for document in docs:
                if callable(document):
                    try:
                        document = await run_async(document())
                    except Exception as error:
                        print(stylize_error(f"Error: {error}"), file=sys.stderr)
                        continue
                for i in range(0, len(document), chunk_size - overlap):
                    chunk = document[i : i + chunk_size]
                    if len(chunk) > 0:
                        print(
                            stylize_faint(f"Vectorize chunk {chunk_index}"),
                            file=sys.stderr,
                        )
                        response = await litellm.aembedding(model=model, input=[chunk])
                        vector = response["data"][0]["embedding"]
                        print(
                            stylize_faint(f"Adding chunk {chunk_index} to db"),
                            file=sys.stderr,
                        )
                        collection.upsert(
                            ids=[f"id{chunk_index}"],
                            embeddings=[vector],
                            documents=[chunk],
                        )
                        chunk_index += 1
        collection = client.get_or_create_collection(vector_db_collection)
        # Generate embedding for the query
        print(stylize_faint("Vectorize query"), file=sys.stderr)
        query_response = await litellm.aembedding(model=model, input=[query])
        print(stylize_faint("Search documents"), file=sys.stderr)
        # Search for the top_k most similar documents
        results = collection.query(
            query_embeddings=query_response["data"][0]["embedding"],
            n_results=max_result_count,
        )
        return json.dumps(results)

    retrieve.__name__ = tool_name
    retrieve.__doc__ = tool_description
    return retrieve


def get_rag_documents(document_dir_path: str) -> Callable[[], list[Callable[[], str]]]:
    def get_documents() -> list[Callable[[], str]]:
        # Walk through the directory
        readers = []
        for root, _, files in os.walk(document_dir_path):
            for file in files:
                file_path = os.path.join(root, file)
                if file_path.lower().endswith(".pdf"):
                    readers.append(_get_pdf_reader(file_path))
                    continue
                readers.append(_get_text_reader(file_path))
        return readers

    return get_documents


def _get_text_reader(file_path: str):
    def read():
        print(stylize_faint(f"Start reading {file_path}"), file=sys.stderr)
        with open(file_path, "r", encoding="utf-8") as f:
            content = f.read()
        print(stylize_faint(f"Complete reading {file_path}"), file=sys.stderr)
        return content

    return read


def _get_pdf_reader(file_path):
    def read():
        import pdfplumber

        print(stylize_faint(f"Start reading {file_path}"), file=sys.stderr)
        contents = []
        with pdfplumber.open(file_path) as pdf:
            for page in pdf.pages:
                contents.append(page.extract_text())
        print(stylize_faint(f"Complete reading {file_path}"), file=sys.stderr)
        return "\n".join(contents)

    return read


def get_rag_reset_db(
    document_dir_path: str, vector_db_path: str = "./chroma"
) -> Callable[[], bool]:
    def should_reset_db() -> bool:
        document_exist = os.path.isdir(document_dir_path)
        if not document_exist:
            raise ValueError(f"Document directory not exists: {document_dir_path}")
        vector_db_exist = os.path.isdir(vector_db_path)
        if not vector_db_exist:
            return True
        document_mtime = _get_most_recent_mtime(document_dir_path)
        vector_db_mtime = _get_most_recent_mtime(vector_db_path)
        return document_mtime > vector_db_mtime

    return should_reset_db


def _get_most_recent_mtime(directory):
    most_recent_mtime = 0
    for root, dirs, files in os.walk(directory):
        # Check mtime for directories
        for name in dirs + files:
            file_path = os.path.join(root, name)
            mtime = os.path.getmtime(file_path)
            if mtime > most_recent_mtime:
                most_recent_mtime = mtime
    return most_recent_mtime



import os
import subprocess
import tempfile
from collections.abc import Callable

from zrb.config import DEFAULT_EDITOR
from zrb.context.any_shared_context import AnySharedContext
from zrb.input.base_input import BaseInput


class TextInput(BaseInput):
    def __init__(
        self,
        name: str,
        description: str | None = None,
        prompt: str | None = None,
        default_str: str | Callable[[AnySharedContext], str] = "",
        auto_render: bool = True,
        allow_empty: bool = False,
        editor: str = DEFAULT_EDITOR,
        extension: str = ".txt",
        comment_start: str | None = None,
        comment_end: str | None = None,
    ):
        super().__init__(
            name=name,
            description=description,
            prompt=prompt,
            default_str=default_str,
            auto_render=auto_render,
            allow_empty=allow_empty,
        )
        self._editor = editor
        self._extension = extension
        self._comment_start = comment_start
        self._comment_end = comment_end

    @property
    def comment_start(self) -> str:
        if self._comment_start is not None:
            return self._comment_start
        if self._extension.lower() in [".py", ".rb", ".sh"]:
            return "# "
        if self._extension.lower() in [".md", ".html"]:
            return "<!-- "
        return "//"

    @property
    def comment_end(self) -> str:
        if self._comment_end is not None:
            return self._comment_end
        if self._extension.lower() in [".md", ".html"]:
            return " -->"
        return ""

    def to_html(self, ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = self._get_default_str(ctx)
        return "\n".join(
            [
                f'<textarea name="{name}" placeholder="{description}">',
                default,
                "</textarea>",
            ]
        )

    def _prompt_cli_str(self, shared_ctx: AnySharedContext) -> str:
        prompt_message = (
            f"{self.comment_start}{super().prompt_message}{self.comment_end}"
        )
        prompt_message_eol = f"{prompt_message}\n"
        default_value = self._get_default_str(shared_ctx)
        with tempfile.NamedTemporaryFile(
            delete=False, suffix=self._extension
        ) as temp_file:
            temp_file_name = temp_file.name
            temp_file.write(prompt_message_eol.encode())
            # Pre-fill with default content
            if default_value:
                temp_file.write(default_value.encode())
            temp_file.flush()
        # Open the editor
        subprocess.call([self._editor, temp_file_name])
        # Read the edited content
        with open(temp_file_name, "r") as temp_file:
            edited_content = temp_file.read()
            parts = [text.strip() for text in edited_content.split(prompt_message, 1)]
            edited_content = "\n".join(parts).lstrip()
        os.remove(temp_file_name)
        return edited_content



import os
import shutil
from collections.abc import Callable

from zrb.attr.type import BoolAttr, StrAttr
from zrb.content_transformer.any_content_transformer import AnyContentTransformer
from zrb.content_transformer.content_transformer import ContentTransformer
from zrb.context.any_context import AnyContext
from zrb.env.any_env import AnyEnv
from zrb.input.any_input import AnyInput
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.util.attr import get_str_attr

TransformConfig = dict[str, str] | Callable[[AnyContext, str], str]


class Scaffolder(BaseTask):
    def __init__(
        self,
        name: str,
        color: int | None = None,
        icon: str | None = None,
        description: str | None = None,
        cli_only: bool = False,
        input: list[AnyInput] | AnyInput | None = None,
        env: list[AnyEnv] | AnyEnv | None = None,
        source_path: StrAttr | None = None,
        render_source_path: bool = True,
        destination_path: StrAttr | None = None,
        render_destination_path: bool = True,
        transform_path: TransformConfig = {},
        transform_content: (
            list[AnyContentTransformer] | AnyContentTransformer | TransformConfig
        ) = [],
        execute_condition: BoolAttr = True,
        retries: int = 2,
        retry_period: float = 0,
        readiness_check: list[AnyTask] | AnyTask | None = None,
        readiness_check_delay: float = 0.5,
        readiness_check_period: float = 5,
        readiness_failure_threshold: int = 1,
        readiness_timeout: int = 60,
        monitor_readiness: bool = False,
        upstream: list[AnyTask] | AnyTask | None = None,
        fallback: list[AnyTask] | AnyTask | None = None,
    ):
        super().__init__(
            name=name,
            color=color,
            icon=icon,
            description=description,
            cli_only=cli_only,
            input=input,
            env=env,
            execute_condition=execute_condition,
            retries=retries,
            retry_period=retry_period,
            readiness_check=readiness_check,
            readiness_check_delay=readiness_check_delay,
            readiness_check_period=readiness_check_period,
            readiness_failure_threshold=readiness_failure_threshold,
            readiness_timeout=readiness_timeout,
            monitor_readiness=monitor_readiness,
            upstream=upstream,
            fallback=fallback,
        )
        self._source_path = source_path
        self._render_source_path = render_source_path
        self._destination_path = destination_path
        self._render_destination_path = render_destination_path
        self._content_transformers = transform_content
        self._path_transformer = transform_path

    def _get_source_path(self, ctx: AnyContext) -> str:
        return get_str_attr(ctx, self._source_path, "", auto_render=True)

    def _get_destination_path(self, ctx: AnyContext) -> str:
        return get_str_attr(ctx, self._destination_path, "", auto_render=True)

    def _get_content_transformers(self) -> list[AnyContentTransformer]:
        if callable(self._content_transformers):
            return [ContentTransformer(match="*", transform=self._content_transformers)]
        if isinstance(self._content_transformers, dict):
            return [ContentTransformer(match="*", transform=self._content_transformers)]
        if isinstance(self._content_transformers, AnyContentTransformer):
            return [self._content_transformers]
        return self._content_transformers

    async def _exec_action(self, ctx: AnyContext):
        source_path = self._get_source_path(ctx)
        destination_path = self._get_destination_path(ctx)
        self._copy_path(ctx, source_path, destination_path)
        transformers = self._get_content_transformers()
        file_paths = self._get_all_file_paths(destination_path)
        for file_path in file_paths:
            for transformer in transformers:
                if transformer.match(ctx, file_path):
                    try:
                        transformer.transform_file(ctx, file_path)
                    except UnicodeDecodeError:
                        pass

    def _copy_path(self, ctx: AnyContext, source_path: str, destination_path: str):
        """
        Copies a directory or file from source_path to destination_path recursively.
        """
        if os.path.isdir(source_path):
            for root, dirs, files in os.walk(source_path):
                rel_root = os.path.relpath(root, source_path)
                dest_dir = os.path.join(
                    destination_path, self._transform_path(ctx, rel_root)
                )
                os.makedirs(dest_dir, exist_ok=True)
                for file_name in files:
                    src_file = os.path.join(root, file_name)
                    dest_file = os.path.join(
                        dest_dir, self._transform_path(ctx, file_name)
                    )
                    shutil.copy2(src_file, dest_file)
                    ctx.log_info(f"Copied and renamed {src_file} to {dest_file}")
        else:
            dest_file = os.path.join(
                destination_path,
                self._transform_path(ctx, os.path.basename(source_path)),
            )
            shutil.copy2(source_path, dest_file)
            ctx.log_info(f"Copied and renamed {source_path} to {dest_file}")

    def _transform_path(self, ctx: AnyContext, file_path: str):
        if callable(self._path_transformer):
            return self._path_transformer(ctx, file_path)
        new_file_path = file_path
        for keyword, replacement in self._path_transformer.items():
            new_file_path = new_file_path.replace(keyword, replacement)
        return new_file_path

    def _get_all_file_paths(self, path):
        """
        Returns a list of absolute file paths for all files in the given path, recursively.
        """
        file_paths = []
        for root, _, files in os.walk(path):
            for file in files:
                absolute_path = os.path.abspath(os.path.join(root, file))
                file_paths.append(absolute_path)
        return file_paths



from __future__ import annotations  # Enables forward references

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Coroutine, TypeVar

from zrb.context.any_context import AnyContext
from zrb.group.any_group import AnyGroup
from zrb.session_state_log.session_state_log import SessionStateLog
from zrb.session_state_logger.any_session_state_logger import AnySessionStateLogger
from zrb.task_status.task_status import TaskStatus

if TYPE_CHECKING:
    from zrb.context import any_shared_context
    from zrb.task import any_task

TAnySession = TypeVar("TAnySession", bound="AnySession")


class AnySession(ABC):
    """Abstract base class for managing task execution and context in a session.

    This class handles task lifecycle management, context retrieval,
    deferred task execution, and data exchange between tasks using
    XCom-like functionality.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Name this session"""
        pass

    @property
    @abstractmethod
    def root_group(self) -> AnyGroup | None:
        """Session root group"""
        pass

    @property
    @abstractmethod
    def task_names(self) -> list[str]:
        """Task names in this session"""
        pass

    @property
    @abstractmethod
    def shared_ctx(self) -> any_shared_context.AnySharedContext:
        """Shared context for this session"""
        pass

    @abstractmethod
    def terminate(self):
        """Terminating session"""
        pass

    @property
    @abstractmethod
    def is_terminated(self) -> bool:
        """Whether session is terminated or not"""
        pass

    @property
    @abstractmethod
    def parent(self) -> TAnySession | None:
        """Parent session"""
        pass

    @abstractmethod
    def task_path(self) -> str:
        """Main task's path"""
        pass

    @property
    @abstractmethod
    def final_result(self) -> Any:
        """Main task's result"""
        pass

    @property
    @abstractmethod
    def state_logger(self) -> AnySessionStateLogger:
        """State logger"""
        pass

    @abstractmethod
    def set_main_task(self, main_task: any_task.AnyTask):
        """Set main task"""
        pass

    @abstractmethod
    def as_state_log(self) -> SessionStateLog:
        pass

    @abstractmethod
    def get_ctx(self, task: any_task.AnyTask) -> AnyContext:
        """Retrieves the context for a specific task.

        Args:
            task (AnyTask): The task for which to retrieve the context.

        Returns:
            AnyContext: The context object specific to the provided task.
        """
        pass

    @abstractmethod
    def defer_monitoring(self, task: any_task.AnyTask, coro: Coroutine):
        """Defers the execution of a task's monitoring coroutine for later processing.

        Args:
            task (AnyTask): The task associated with the coroutine.
            coro (Coroutine): The monitoring coroutine to defer.
        """
        pass

    @abstractmethod
    def defer_action(self, task: any_task.AnyTask, coro: Coroutine):
        """Defers the execution of a task's coroutine for later processing.

        Args:
            task (AnyTask): The task associated with the coroutine.
            coro (Coroutine): The coroutine to defer.
        """
        pass

    @abstractmethod
    def defer_coro(self, coro: Coroutine):
        """Defers the execution of a coroutine for later processing.

        Args:
            coro (Coroutine): The coroutine to defer.
        """
        pass

    @abstractmethod
    async def wait_deferred(self):
        """Asynchronously waits for all deffered coroutines to complete"""
        pass

    @abstractmethod
    def register_task(self, task: any_task.AnyTask):
        """Registers a new task in the session.

        Args:
            task (AnyTask): The task to register in the session.
        """
        pass

    @abstractmethod
    def get_root_tasks(self, task: any_task.AnyTask) -> list[any_task.AnyTask]:
        """Retrieves the list of root tasks that should be executed first
        to run the given task.

        Args:
            task (AnyTask): The current task.

        Returns:
            list[AnyTask]: A list of root tasks.
        """
        pass

    @abstractmethod
    def get_next_tasks(self, task: any_task.AnyTask) -> list[any_task.AnyTask]:
        """Retrieves the list of tasks that should be executed after the given task.

        Args:
            task (AnyTask): The current task.

        Returns:
            list[AnyTask]: A list of tasks that should be executed next.
        """
        pass

    @abstractmethod
    def get_task_status(self, task: any_task.AnyTask) -> TaskStatus:
        """Get tasks' status.

        Args:
            task (AnyTask): The task to mark as started.

        Returns:
            TaskStatus: Task status object
        """
        pass

    @abstractmethod
    def is_allowed_to_run(self, task: any_task.AnyTask):
        """Determines if the specified task is allowed to run based on its current state.

        Args:
            task (AnyTask): The task to check.

        Returns:
            bool: True if the task is allowed to run, False otherwise.
        """
        pass



from zrb.attr.type import StrAttr
from zrb.context.any_shared_context import AnySharedContext
from zrb.input.base_input import BaseInput
from zrb.util.string.conversion import to_boolean


class BoolInput(BaseInput):
    def __init__(
        self,
        name: str,
        description: str | None = None,
        prompt: str | None = None,
        default_str: StrAttr = "False",
        auto_render: bool = True,
        allow_empty: bool = False,
    ):
        super().__init__(
            name=name,
            description=description,
            prompt=prompt,
            default_str=default_str,
            auto_render=auto_render,
            allow_empty=allow_empty,
        )

    def to_html(self, ctx: AnySharedContext) -> str:
        name = self.name
        description = self.description
        default = to_boolean(self._get_default_str(ctx))
        selected_true = "selected" if default else ""
        selected_false = "selected" if not default else ""
        return "\n".join(
            [
                f'<select name="{name}" placeholder="{description}">',
                f'<option value="true" {selected_true}>true</option>',
                f'<option value="false" {selected_false}>false</option>',
                "</select>",
            ]
        )

    def _parse_str_value(self, str_value: str) -> bool:
        return to_boolean(str_value)



from fastapp_template.common.db_repository import BaseDBRepository
from fastapp_template.common.error import NotFoundError
from fastapp_template.module.auth.service.user.repository.repository import (
    UserRepository,
)
from fastapp_template.schema.user import User, UserCreate, UserResponse, UserUpdate
from passlib.context import CryptContext
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel import Session, select

# Password hashing context
pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def hash_password(password: str) -> str:
    return pwd_context.hash(password)


class UserDBRepository(
    BaseDBRepository[User, UserResponse, UserCreate, UserUpdate], UserRepository
):
    db_model = User
    response_model = UserResponse
    create_model = UserCreate
    update_model = UserUpdate
    entity_name = "user"
    column_preprocessors = {"password": hash_password}

    async def get_by_credentials(self, username: str, password: str) -> UserResponse:
        statement = select(User).where(User.username == username)
        if self.is_async:
            async with AsyncSession(self.engine) as session:
                user = await session.exec(statement).first
        else:
            with Session(self.engine) as session:
                user = session.exec(statement).first()
        if not user or not pwd_context.verify(password, user.hashed_password):
            raise NotFoundError(f"{self.entity_name} not found")
        return self._to_response(user)



from typing import Dict

from fastapi import HTTPException


class NotFoundError(HTTPException):
    def __init__(self, message: str, headers: Dict[str, str] | None = None) -> None:
        super().__init__(404, {"message": message}, headers)



import os
import subprocess

from pydantic import BaseModel


class SingleSubTreeConfig(BaseModel):
    repo_url: str
    branch: str
    prefix: str


class SubTreeConfig(BaseModel):
    data: dict[str, SingleSubTreeConfig]


def load_config(repo_dir: str) -> SubTreeConfig:
    file_path = os.path.join(repo_dir, "subtrees.json")
    if not os.path.exists(file_path):
        return SubTreeConfig(data={})
    with open(file_path, "r") as f:
        return SubTreeConfig.model_validate_json(f.read())


def save_config(repo_dir: str, config: SubTreeConfig):
    file_path = os.path.join(repo_dir, "subtrees.json")
    with open(file_path, "w") as f:
        f.write(config.model_dump_json(indent=2))


def add_subtree(repo_dir: str, name: str, repo_url: str, branch: str, prefix: str):
    config = load_config()
    if os.path.isdir(prefix):
        raise ValueError(f"Directory exists: {prefix}")
    if name in config.data:
        raise ValueError(f"Subtree config already exists: {name}")
    try:
        subprocess.run(
            ["git", "subtree", "add", "--prefix", prefix, repo_url, branch],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)
    config.data[name] = SingleSubTreeConfig(
        repo_url=repo_url, branch=branch, prefix=prefix
    )
    save_config(repo_dir, config)


def pull_subtree(repo_dir: str, prefix: str, repo_url: str, branch: str):
    try:
        subprocess.run(
            [
                "git",
                "subtree",
                "pull",
                "--prefix",
                prefix,
                repo_url,
                branch,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)


def push_subtree(repo_dir: str, prefix: str, repo_url: str, branch: str):
    try:
        subprocess.run(
            [
                "git",
                "subtree",
                "push",
                "--prefix",
                prefix,
                repo_url,
                branch,
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=repo_dir,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(e.stderr or e.stdout)



from fastapp_template.config import APP_COMMUNICATION
from fastapp_template.module.auth.client.api_client import APIClient
from fastapp_template.module.auth.client.base_client import BaseClient
from fastapp_template.module.auth.client.direct_client import DirectClient

if APP_COMMUNICATION == "direct":
    client: BaseClient = DirectClient()
elif APP_COMMUNICATION == "api":
    client: BaseClient = APIClient()
